#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""
import numpy as np
import biosteam as bst
from biosteam import SystemFactory
from biosteam import main_flowsheet as f
from .bagasse import (
    create_bagasse_drying_system,
    create_bagasse_pelleting_system,
)
from .. import streams as s

__all__ = (
    'create_feedstock_handling_system',
    'create_juicing_system_up_to_clarification',
    'create_juicing_system_without_treatment',
    'create_juicing_system',
)

@SystemFactory(
    ID='feedstock_handling_sys',
    ins=[s.sugarcane],
    outs=[s.shredded_cane]
)
def create_feedstock_handling_system(ins, outs):
    sugarcane, = ins
    shreaded_cane, = outs
    U101 = bst.ConveyingBelt('U101', sugarcane)
    U102 = bst.MagneticSeparator('U102', U101-0)
    U103 = bst.Shredder('U103', U102-0, shreaded_cane)

@SystemFactory(
    ID='juicing_sys',
    ins=[s.sugarcane],
    outs=[s.untreated_juice, s.bagasse]
)
def create_juicing_system_without_treatment(ins, outs, pellet_bagasse=None,
                                            dry_bagasse=None):
    if pellet_bagasse is None: pellet_bagasse = False
    if dry_bagasse is None: dry_bagasse = False
    
    ### Streams ###
    sugarcane, = ins
    untreated_juice, bagasse = outs
    
    imbibition_water = bst.Stream('imbibition_water',
                                  Water=87023.35, units='kg/hr',
                                  T = 350.15)
    
    # Finely crush lipid cane
    U201 = bst.CrushingMill('U201',
                              ins=(sugarcane, ''),
                              split=dict(Ash=0.92,
                                         Cellulose=0.92,
                                         Glucose=0.04,
                                         Hemicellulose=0.92,
                                         Lignin=0.92,
                                         Sucrose=0.04,
                                         Solids=1),
                              moisture_content=0.5)
    
    @U201.add_specification(run=True)
    def update_imbibition_water():
        feed = U201.ins[0]
        moisture = feed.imass['Water']
        dry = feed.F_mass - moisture
        imbibition_water.imass['Water'] = max(2.68333333 * dry - moisture, 0)
    
    U202 = bst.ConveyingBelt('U202', U201-0, [''] if (pellet_bagasse or dry_bagasse) else [bagasse])
    
    if pellet_bagasse:
        create_bagasse_pelleting_system(None, ins=U202-0, outs=bagasse, mockup=True)
    elif dry_bagasse:
        create_bagasse_drying_system(None, ins=U202-0, outs=bagasse, mockup=True)
    
    # Mix in water
    M201 = bst.Mixer('M201', ('', imbibition_water), 1-U201)
    
    # Screen out fibers
    S201 = bst.VibratingScreen('S201', U201-1, ('', 0-M201),
                                 split=dict(Ash=0.35,
                                            Cellulose=0.35,
                                            Glucose=0.88,
                                            Hemicellulose=0.35,
                                            Lignin=0.35,
                                            Solids=0,
                                            Sucrose=0.88,
                                            Water=0.88))
    
    # Store juice before treatment
    T202 = bst.StorageTank('T202', S201-0, untreated_juice,
                             tau=4, vessel_material='Carbon steel')
    
    if 'Lipid' in bst.settings.chemicals:
        u = f.unit
        u.U201.isplit['Lipid'] = 0.90 # Crushing mill
        u.S201.isplit['Lipid'] = 1.0 # Fiber screener #1


@SystemFactory(
    ID='juicing_sys',
    ins=[s.sugarcane, s.H3PO4, s.lime, s.polymer],
    outs=[s.clarified_juice, s.bagasse]
)
def create_juicing_system_up_to_clarification(ins, outs, pellet_bagasse=None,
                                              dry_bagasse=None):
    
    ### Streams ###
    sugarcane, H3PO4, lime, polymer = ins
    clarified_juice, bagasse = outs
    
    rvf_wash_water = bst.Stream('rvf_wash_water',
                                Water=16770, units='kg/hr',
                                T=363.15)  # to C202
    
    juicing_sys_without_treatment = create_juicing_system_without_treatment(
        ins=(sugarcane,),
        outs=('', bagasse),
        mockup=True,
        pellet_bagasse=pellet_bagasse,
        dry_bagasse=dry_bagasse,
    )
    U201 = f.unit.U201
    
    # Heat up before adding acid
    H201 = bst.HXutility('H201', juicing_sys_without_treatment-0, T=343.15)
    
    # Mix in acid
    T203 = bst.MixTank('T203', (H201-0, H3PO4))
    
    # Pump acid solution
    P201 = bst.Pump('P201', T203-0)
    
    # Mix lime solution
    T204 = bst.MixTank('T204', lime, tau=0.10)
    
    # Blend acid lipid solution with lime
    T205 = bst.MixTank('T205', (P201-0, T204-0), tau=0.10)
    P202 = bst.Pump('P202', T205-0)
    
    # Mix recycle
    M202 = bst.Mixer('M202', (P202-0, ''))
    
    # Heat before adding flocculant
    H202 = bst.HXutility('H202', M202-0, T=372.15)
    
    # Mix in flocculant
    T206 = bst.MixTank('T206', (H202-0, polymer))
    T206.tau = 0.10
    
    # Separate residual solids
    C201 = bst.Clarifier('C201', T206-0, (clarified_juice, ''),
                           split=dict(Ash=0,
                                      CaO=0,
                                      Cellulose=0,
                                      Flocculant=0.522,
                                      Glucose=0.522,
                                      Hemicellulose=0,
                                      Lignin=0,
                                      H3PO4=0.522,
                                      Sucrose=0.522,
                                      Water=0.522))
    
    # Remove solids as filter cake
    C202 = bst.RVF('C202', (C201-1, rvf_wash_water), ('filter_cake', ''),
                     moisture_content=0.80,
                     split=dict(Ash=0.85,
                                CaO=0.85,
                                Cellulose=0.85,
                                Glucose=0.01,
                                Hemicellulose=0.85,
                                Lignin=0.85,
                                Sucrose=0.01))
    P203 = bst.Pump('P203', C202-1, 1-M202)
    
    ### Process specifications ###
    
    # Specifications dependent on lipid cane flow rate
    @U201.add_specification(run=True)
    def correct_flows():
        feed = U201.ins[0]
        other_liquids = feed.imass['Lipid'] if 'Lipid' in feed.chemicals else 0.
        F_mass = (feed.F_mass - feed.imass['Water'] - other_liquids) / 0.7
        # correct lime, phosphoric acid, and imbibition water
        lime.imass['CaO', 'Water'] = 0.001 * F_mass * np.array([0.046, 0.954])
        H3PO4.imass['H3PO4', 'Water'] = 0.00025 * F_mass
    
    # Specifications within a system
    @P202.add_specification
    def correct_wash_water():
        P202._run()
        solids = P202.outs[0].imol['Ash', 'CaO', 'Cellulose',
                                   'Hemicellulose', 'Lignin'].sum()
        rvf_wash_water.imol['Water'] = 0.0574 * solids
        
    if 'Lipid' in bst.settings.chemicals:
        u = f.unit
        u.C201.isplit['Lipid'] = 0.99 # Clarifier
    

@SystemFactory(
    ID='juicing_sys',
    ins=[s.sugarcane, s.H3PO4, s.lime, s.polymer],
    outs=[s.screened_juice, s.bagasse, s.fiber_fines]
)          
def create_juicing_system(ins, outs, pellet_bagasse=None, dry_bagasse=None):
    screened_juice, bagasse, fiber_fines = outs
    sys = create_juicing_system_up_to_clarification(
        None, ins, ['', bagasse],
        mockup=True,
        pellet_bagasse=pellet_bagasse,
        dry_bagasse=dry_bagasse,
    )
    # Screen out small fibers from sugar stream
    S202 = bst.VibratingScreen('S202', sys-0, (screened_juice, fiber_fines),
                                 split=dict(Ash=0.998,
                                            CaO=0.998,
                                            Cellulose=0.0,
                                            Flocculant=0.0,
                                            Glucose=0.998,
                                            Hemicellulose=0.0,
                                            Lignin=0.0,
                                            H3PO4=1.0,
                                            Sucrose=0.998,
                                            Water=0.998))
    S202.mesh_opening = 2
    if 'Lipid' in bst.settings.chemicals:
        u = f.unit
        u.S202.isplit['Lipid'] = 0.999  # Fiber screener #2



# -*- coding: utf-8 -*-
"""
Created on Thu Nov  4 14:39:10 2021

@author: yrc2
"""
from . import feature_mockups as f
from .parse_configuration import parse_configuration, Configuration, ConfigurationComparison
from warnings import warn
from thermosteam.utils import roundsigfigs
import os
import pandas as pd
import numpy as np
import biosteam as bst

__all__ = (
    'images_folder',
    'results_folder',
    'spearman_file',
    'monte_carlo_file',
    'autoload_file_name',
    'get_monte_carlo_across_oil_content',
    'get_monte_carlo',
    'get_line_monte_carlo',
    'montecarlo_results',
    'montecarlo_results_short',
    'montecarlo_results_feedstock_comparison',
    'montecarlo_results_configuration_comparison',
    'montecarlo_results_agile_comparison',
    'montecarlo_results_crude_comparison',
    'get_minimum_GWP_reduction',
)

results_folder = os.path.join(os.path.dirname(__file__), 'results')
images_folder = os.path.join(os.path.dirname(__file__), 'images')

# %% Load simulation data

def spearman_file(name, line=None):
    number, agile, energy_cane = parse_configuration(name)
    filename = f'oilcane_spearman_{number}'
    if agile: filename += '_agile'
    if line: filename += '_' + line
    if energy_cane: filename += '_energy_cane'
    filename += '.xlsx'
    return os.path.join(results_folder, filename)

def monte_carlo_file(name, across_lines=False, across_oil_content=None, line=None, extention='xlsx'):
    number, agile, energycane = parse_configuration(name)
    filename = f'oilcane_monte_carlo_{number}'
    if agile: filename += '_agile'
    if line: filename += '_' + line
    if energycane: filename += '_energycane'
    if across_lines: filename += '_across_lines'
    if across_oil_content: 
        if isinstance(across_oil_content, str):
            filename += f"_{across_oil_content.replace(' ', '_')}"
        else:
            filename += '_across_oil_content'
    filename += '.' + extention
    return os.path.join(results_folder, filename)

def autoload_file_name(name, line=None):
    filename = str(name).replace('*', '_agile')
    if line: filename += '_' + line
    return os.path.join(results_folder, filename)

def get_monte_carlo_across_oil_content(name, metric, derivative=False):
    key = parse_configuration(name)
    if isinstance(key, Configuration):
        df = pd.read_excel(
            monte_carlo_file(key, True),
            sheet_name=metric if isinstance(metric, str) else metric.short_description,
            index_col=0
        )
    elif isinstance(key, ConfigurationComparison):
        df = (
            get_monte_carlo_across_oil_content(key.a, metric)
            - get_monte_carlo_across_oil_content(key.b, metric)
        )
    else:
        raise Exception('unknown error')
    if derivative: 
        arr = np.diff(df.values) / np.diff(df.columns.values) / 100.
    else:
        arr = df.values
    return arr
        
def get_monte_carlo_key(index, dct, with_units=False):
    key = index[1] if with_units else index[1].split(' [')[0]
    if key in dct: key = f'{key}, {index[0]}'
    return key

def get_line_monte_carlo(line, name, feature, cache={}):
    configuration = parse_configuration(name)
    key = (*configuration, feature.short_description)
    if isinstance(configuration, Configuration):
        if key in cache:
            df = cache[key]
        else:
            file = monte_carlo_file(configuration, across_lines=True)
            cache[key] = df = pd.read_excel(file, header=[0, 1], index_col=[0], sheet_name=feature.short_description)
        mc = df[line]
    elif isinstance(configuration, ConfigurationComparison):
        raise ValueError('name cannot be a configuration comparison')
    else:
        raise Exception('unknown error')
    mc = mc.dropna(how='all', axis=0)
    return mc

def get_monte_carlo(name, features=None, line=None, cache={}):
    key = parse_configuration(name)
    if isinstance(key, Configuration):
        if key in cache:
            df = cache[key]
        else:
            file = monte_carlo_file(key, line=line)
            cache[key] = df = pd.read_excel(file, header=[0, 1], index_col=[0])
        if features is None:
            mc = df
        elif isinstance(features, bst.Feature):
            mc = df[features.index]
        else:
            mc = df[[i.index for i in features]]
    elif isinstance(key, ConfigurationComparison):
        if features is None:
            features = (
                *f.tea_monte_carlo_metric_mockups, 
                *f.tea_monte_carlo_derivative_metric_mockups,
                *f.lca_monte_carlo_metric_mockups, 
                *f.lca_monte_carlo_derivative_metric_mockups,
                f.net_energy_production,
                f.GWP_ethanol_displacement,
                f.GWP_ethanol_allocation,
            )
        if isinstance(features, bst.Feature):
            index = features.index
        else:
            index = [i.index for i in features]
        df_a = get_monte_carlo(key.a, line=line)[index]
        df_b = get_monte_carlo(key.b, line=line)[index]
        row_a = df_a.shape[0]
        row_b = df_b.shape[0]
        try:
            assert row_a == row_b, "shape mismatch"
        except:
            breakpoint()
        mc = df_a - df_b
    else:
        raise Exception('unknown error')
    mc = mc.dropna(how='all', axis=0)
    return mc


def montecarlo_results(with_units=False):
    results = {}    
    for name in ('S1', 'O1', 'S2', 'O2', 'S1*', 'O1*', 'S2*', 'O2*', 'O3', 'O4', 
                 'O1 - S1', 'O2 - S2',  'O2 - O1', 'O1* - O1',  'O2* - O2'):
        try: 
            df = get_monte_carlo(name)
        except:
            warn(f'could not load {name}', RuntimeWarning)
            continue
        results[name] = dct = {}
        # if name in ('O1', 'O2'):
        #     index = f.ethanol_over_biodiesel.index
        #     key = get_monte_carlo_key(index, dct, with_units)
        #     data = df[f.ethanol_production.index].values / df[f.biodiesel_production.index].values
        #     q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
        #     dct[key] = {
        #         'mean': np.mean(data),
        #         'std': np.std(data),
        #         'q05': q05,
        #         'q25': q25,
        #         'q50': q50,
        #         'q75': q75,
        #         'q95': q95,
        #     }
        for metric in (*f.tea_monte_carlo_metric_mockups, *f.tea_monte_carlo_derivative_metric_mockups,
                       *f.lca_monte_carlo_metric_mockups, *f.lca_monte_carlo_derivative_metric_mockups,
                       f.GWP_ethanol_displacement, f.GWP_ethanol_allocation):
            index = metric.index
            data = df[index].values
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            key = get_monte_carlo_key(index, dct, with_units)
            dct[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
    try:
        df_O2O1 = get_monte_carlo('O2 - O1')
        df_O1 = get_monte_carlo('O1')
    except:
        warn('could not load O2 - O1', RuntimeWarning)
    else:
        results['(O2 - O1) / O1'] = relative_results = {}
        for metric in (f.biodiesel_production, f.ethanol_production):
            index = metric.index
            key = index[1] if with_units else index[1].split(' [')[0]
            data = (df_O2O1[index].values / df_O1[index].values)
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            relative_results[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
    return results

# %% Monte carlo values for manuscript

def get_minimum_GWP_reduction():
    L_per_gal = 3.7854
    gal_per_GGE_ethanol = 1.5
    L_per_GGE_ethanol = L_per_gal * gal_per_GGE_ethanol
    gal_per_GGE_biodiesel = 0.9536
    L_per_GGE_biodiesel = L_per_gal * gal_per_GGE_biodiesel
    GWP_ethanol_economic = GWP_biodiesel_economic = 0
    GWP_ethanol_energy = GWP_biodiesel_energy = 0
    GWP_ethanol_economic_index = f.GWP_ethanol.index
    GWP_ethanol_energy_index = f.GWP_ethanol_allocation.index
    GWP_biodiesel_economic_index = f.GWP_biodiesel.index
    GWP_biodiesel_energy_index = f.GWP_biodiesel_allocation.index
    # https://www.epa.gov/sites/default/files/2016-07/documents/select-ghg-results-table-v1.pdf
    GWP_ethanol_dist_and_use = 3.7 * 114000 / 1e6 # kgCO2e per GGE
    GWP_biodiesel_dist_and_use = 3.4 * 114000 / 1e6 # kgCO2e per GGE
    for name in ('O1', 'O2'): 
        df = get_monte_carlo(name)
        GWP_ethanol_economic = max(GWP_ethanol_economic, df[GWP_ethanol_economic_index].max())
        GWP_ethanol_energy = max(GWP_ethanol_energy, df[GWP_ethanol_energy_index].max())
        GWP_biodiesel_economic = max(GWP_biodiesel_economic, df[GWP_biodiesel_economic_index].max())
        GWP_biodiesel_energy = max(GWP_biodiesel_energy, df[GWP_biodiesel_energy_index].max())
    # https://www.epa.gov/sites/default/files/2016-07/documents/select-ghg-results-table-v1.pdf
    GWP_ethanol_economic *= L_per_GGE_ethanol # per L to per GGE
    GWP_ethanol_energy *= L_per_GGE_ethanol # per L to per GGE
    GWP_biodiesel_economic *= L_per_GGE_biodiesel # per L to per GGE
    GWP_biodiesel_energy *= L_per_GGE_biodiesel # per L to per GGE
    GWP_diesel = 11.058 # kgCO2e per GGE
    GWP_gasoline = 11.1948 # kgCO2e per GGE
    return {
        'Ethanol-economic': 100 * (1 - (GWP_ethanol_economic + GWP_ethanol_dist_and_use) / GWP_gasoline),
        'Ethanol-energy': 100 * (1 - (GWP_ethanol_energy + GWP_ethanol_dist_and_use)/ GWP_gasoline),
        'Biodiesel-economic': 100 * (1 - (GWP_biodiesel_economic + GWP_biodiesel_dist_and_use)/ GWP_diesel),
        'Biodiesel-energy': 100 * (1 - (GWP_biodiesel_energy + GWP_biodiesel_dist_and_use) / GWP_diesel),
    }

def montecarlo_results_short(names, metrics=None, derivative=None):
    if metrics is None:
        if derivative:
            metrics = [
                f.MFPP_derivative, f.TCI_derivative, f.ethanol_production_derivative, f.biodiesel_production_derivative, 
                f.electricity_production_derivative, f.natural_gas_consumption_derivative, 
                f.GWP_ethanol_derivative, 
            ]
        else:
            metrics = [
                f.MFPP, f.TCI, f.ethanol_production, f.biodiesel_production, 
                f.electricity_production, f.natural_gas_consumption, f.GWP_ethanol_displacement, 
                f.GWP_ethanol, f.GWP_ethanol, f.GWP_biodiesel, 
                f.net_energy_production,
            ]
    results = {}
    for name in names:
        df = get_monte_carlo(name)
        results[name] = dct = {}
        for metric in metrics:
            index = metric.index
            try:
                data = df[index].values
            except:
                breakpoint()
            q05, q50, q95 = roundsigfigs(np.percentile(data, [5, 50, 95], axis=0), 3)
            key = get_monte_carlo_key(index, dct, False)
            if q50 < 0:
                dct[key] = f"{-q50} [{-q95}, {-q05}] -negative-"
            else:
                dct[key] = f"{q50} [{q05}, {q95}]"
        if name == 'O2 - O1':
            df_O2O1 = df
            df_O1 = get_monte_carlo('O1')
            for metric in (f.biodiesel_production, f.ethanol_production):
                index = metric.index
                key = f"% {get_monte_carlo_key(index, {}, False)} increase"
                data = 100. * (df_O2O1[index].values / df_O1[index].values)
                q05, q50, q95 = roundsigfigs(np.percentile(data, [5, 50, 95], axis=0), 3)
                if q50 < 0:
                    dct[key] = f"{-q50} [{-q95}, {-q05}] -negative-"
                else:
                    dct[key] = f"{q50} [{q05}, {q95}]"
    return results

def montecarlo_results_feedstock_comparison():
    return montecarlo_results_short(
        names = [
            'O1 - S1',
            'O2 - S2',
        ],
    )
    
def montecarlo_results_configuration_comparison():
    return montecarlo_results_short(
        names = [
            'O2 - O1',
            'O1 - O3',
            'O2 - O4',
        ],
    )

def montecarlo_results_agile_comparison():
    return montecarlo_results_short(
        names = [
            'O1* - O1',
            'O2* - O2',
        ],
    )

def montecarlo_results_crude_comparison():
    return montecarlo_results_short(
        names = [
            'O1 - O3',
            'O2 - O4',
        ],
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst
from biosteam import SystemFactory, F
from ... import streams as s
from biorefineries.ethanol import (
    create_ethanol_purification_system,
)
from ..fermentation import create_sucrose_fermentation_system
from ..juicing import create_feedstock_handling_system, create_juicing_system

__all__ = (
    'create_sucrose_to_ethanol_system',
    'create_sugarcane_to_ethanol_system',
)

# %% Sugarcane to ethanol

@SystemFactory(
    ID='sucrose_to_ethanol_sys',
    ins=[s.screened_juice, s.denaturant],
    outs=[s.ethanol, s.stillage, s.recycle_process_water, s.evaporator_condensate]
)
def create_sucrose_to_ethanol_system(ins, outs, add_urea=False):
    screened_juice, denaturant = ins
    ethanol, stillage, recycle_process_water, evaporator_condensate = outs
    
    beer = bst.Stream()
    
    create_sucrose_fermentation_system(
        ins=screened_juice,
        outs=[beer, evaporator_condensate],
        mockup=True,
        add_urea=add_urea,
    )
    create_ethanol_purification_system(
        ins=[beer, denaturant], 
        outs=[ethanol, stillage, recycle_process_water],
        mockup=True
    )

@SystemFactory(
    ID='sugarcane_sys', 
    ins=[s.sugarcane, s.H3PO4, s.lime, s.polymer, s.denaturant], 
    outs=[s.ethanol, s.vinasse, s.fiber_fines, s.emissions, s.ash_disposal]
)
def create_sugarcane_to_ethanol_system(ins, outs, 
                                       use_area_convention=False,
                                       pellet_bagasse=None,
                                       dry_bagasse=None,
                                       WWT_kwargs=None):
    sugarcane, H3PO4, lime, polymer, denaturant = ins
    ethanol, vinasse, fiber_fines, emissions, ash_disposal = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        area=100 if use_area_convention else None,
        ins=[sugarcane],
        outs=[''],
        mockup=True,
    )
    juicing_sys = create_juicing_system(
        area=200 if use_area_convention else None,
        ins=[feedstock_handling_sys-0, H3PO4, lime, polymer],
        outs=['juice', 'bagasse', fiber_fines],
        pellet_bagasse=pellet_bagasse,
        dry_bagasse=dry_bagasse,
        mockup=True
    )
    ethanol_production_sys, edct = create_sucrose_to_ethanol_system(
        area=300 if use_area_convention else None,
        udct=True,
        ins=(juicing_sys-0, denaturant), outs=(ethanol, vinasse),
        mockup=True
    )
    M305 = bst.Mixer(400 if use_area_convention else 'M305', 
        ins=ethanol_production_sys-[2, 3],
    )
    
    ### Facilities ###    
    CT = bst.CoolingTower(500 if use_area_convention else 'CT')
    CWP = bst.ChilledWaterPackage(500 if use_area_convention else 'CWP')
    
    F301 = edct['F301']
    D303 = edct['D303']
    HXN = bst.HeatExchangerNetwork(600 if use_area_convention else 'HXN',
                                   units=[F301, D303.condenser])
    MX1 = bst.Mixer(400, [juicing_sys-1])
    if WWT_kwargs:
        outs.remove(vinasse)
        wastewater_treatment_sys = bst.create_wastewater_treatment_system(
             ins=[vinasse], mockup=True, area=700, **WWT_kwargs
        )
        treated_water = wastewater_treatment_sys.get_outlet('RO_treated_water')
        sludge = wastewater_treatment_sys.get_outlet('sludge')
        biogas = wastewater_treatment_sys.get_outlet('biogas')
        MX1.ins.append(sludge)
    else:
        treated_water = ''
        biogas = ''
    
    BT = bst.BoilerTurbogenerator(400 if use_area_convention else 'BT',
        (MX1-0, biogas, 'boiler_makeup_water', 'natural_gas', '', ''),
        outs=(emissions, 'rejected_water_and_blowdown', ash_disposal),
        boiler_efficiency=0.80,
        turbogenerator_efficiency=0.85
    )
    PWC = bst.ProcessWaterCenter(500 if use_area_convention else 'PWC',
                                 ins=[treated_water, 'makeup_RO_water', M305-0, ''],
                                 process_water_price=0.000254)
    
    # if vinasse_to_wastewater:
    #     plant_air = bst.Stream('plant_air', N2=83333, units='kg/hr')
    #     ADP = bst.facilities.AirDistributionPackage('ADP', plant_air)
    #     @ADP.add_specification(run=True)
    #     def adjust_plant_air():
    #         plant_air.imass['N2'] = 0.8 * feedstock_handling_sys.ins[0].F_mass
            
    #     wastewater_treatment_sys = bst.create_wastewater_treatment_system(
    #         ins=[vinasse],
    #         mockup=True,
    #     )


# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst

__all__ = (
    'load_process_settings',
)

def load_process_settings():
    bst.process_tools.default_utilities()
    settings = bst.settings 
    settings.CEPCI = 525.4
    settings.electricity_price = 0.0572 # USD/kWh
    lps = settings.get_heating_agent('low_pressure_steam')
    lps.heat_transfer_efficiency = 0.90
    lps.T = 529.2
    lps.P = 44e5
    lps.regeneration_price = 0.
    cw = settings.get_cooling_agent('cooling_water')
    cw.T = 28 + 273.15
    cw.T_limit = cw.T + 9
    cw.regeneration_price = 0.
    settings.get_cooling_agent('chilled_water').heat_transfer_price = 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>,
#                      Sarang Bhagwat <sarangb2@illinois.edu>,
#                      Yoel Cortes-Pena <yoelcortes@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

import thermosteam as tmo, biosteam as bst
auom = tmo.units_of_measure.AbsoluteUnitsOfMeasure

# Using BioSTEAM's default values
# (about the same as the previous dict, only as integers)
CEPCI = bst.design_tools.CEPCI_by_year
# # Legacy Chemical Engineering Plant Cost Index from Chemical Engineering Magazine
# # (https://www.chemengonline.com/the-magazine/)
# CEPCI = {1997: 386.5,
#          1998: 389.5,
#          2007: 525.4,
#          2009: 521.9,
#          2010: 550.8,
#          2011: 585.7,
#          2012: 584.6,
#          2013: 567.3,
#          2014: 576.1,
#          2016: 541.7}


# %%

from ._chemicals import *
from .utils import *
from ._process_settings import *
from ._processes import *
from ._tea import *
from .systems import *

from . import (
    _chemicals,
    utils,
    _process_settings,
    _units,
    _processes,
    _tea,
    systems,
    )

_chemicals_loaded = False
_system_loaded = False

def load(kind='SSCF', print_results=True, **sys_kwdct):
    if not _chemicals_loaded: _load_chemicals()
    kind = kind.upper()
    if not kind in ('SSCF', 'SHF'):
        raise ValueError(f'kind can only be "SSCF" or "SHF", not "{kind}".')
    _load_system(kind, **sys_kwdct)
    dct = globals()
    dct.update(flowsheet.to_dict())
    if print_results: simulate_and_print(flowsheet=flowsheet)

def _load_chemicals():
    global chemicals, _chemicals_loaded
    chemicals = create_chemicals()
    _chemicals_loaded = True

def _load_system(kind='SSCF', **sys_kwdct):
    global flowsheet, funcs, lactic_sys, lactic_tea
    flowsheet = bst.Flowsheet('lactic')
    bst.main_flowsheet.set_flowsheet(flowsheet)
    bst.settings.set_thermo(chemicals)
    load_process_settings()
    lactic_sys, groups = create_system(kind=kind, return_groups=True, flowsheet=flowsheet, **sys_kwdct)
    global Area100, Area200, Area300, Area400, Area500, Area600, CHP
    Area100, Area200, Area300, Area400, Area500, HXN, BT, CT, Area600 = groups
    CHP = BT
    lactic_tea = create_tea(flowsheet=flowsheet)
    funcs = create_funcs(lactic_tea=lactic_tea, flowsheet=flowsheet)
    global _system_loaded
    _system_loaded = True


# %%

# =============================================================================
# Useful functions for summarizing results and considering alternative process
# decision variables
# =============================================================================

global simulate_and_print
def simulate_and_print(funcs=None, lactic_tea=None, flowsheet=None):
    funcs = funcs or create_funcs(lactic_tea=lactic_tea, flowsheet=flowsheet)
    MPSP = funcs['simulate_get_MPSP']()
    GWP = funcs['get_GWP']()
    FEC = funcs['get_FEC']()
    print('\n---------- Simulation Results ----------')
    print(f'MPSP is ${MPSP:.3f}/kg')
    print(f'GWP is {GWP:.3f} kg CO2-eq/kg lactic acid')
    print(f'FEC is {FEC:.2f} MJ/kg lactic acid')
    print('------------------------------------------\n')


global simulate_fermentation_improvement
def simulate_fermentation_improvement(funcs=None, lactic_tea=None, flowsheet=None):
    flowsheet = flowsheet or bst.main_flowsheet
    funcs = funcs or create_funcs(lactic_tea=lactic_tea, flowsheet=flowsheet)
    u = flowsheet.unit
    sys = flowsheet.system.lactic_sys
    sys.simulate()
    R301_X = u.R301.cofermentation_rxns.X
    R302_X = u.R302.cofermentation_rxns.X
    u.R301.target_yield = 0.95
    R301_X[0] = R301_X[3] = 0.95
    R301_X[1] = R301_X[4] = 0
    R302_X[1] = R302_X[4] = 0
    simulate_and_print(funcs=funcs, lactic_tea=sys.TEA, flowsheet=flowsheet)


global simulate_separation_improvement
def simulate_separation_improvement(funcs=None, lactic_tea=None, flowsheet=None):
    flowsheet = flowsheet or bst.main_flowsheet
    funcs = funcs or create_funcs(lactic_tea=lactic_tea, flowsheet=flowsheet)
    u = flowsheet.unit
    sys = flowsheet.system.lactic_sys
    sys.simulate()
    u.R402.X_factor = 0.9/u.R402.esterification_rxns.X[0]
    u.R403.hydrolysis_rxns.X[:] = 0.9
    simulate_and_print(funcs=funcs, lactic_tea=sys.TEA, flowsheet=flowsheet)


global simulate_operating_improvement
def simulate_operating_improvement(funcs=None, lactic_tea=None, flowsheet=None):
    flowsheet = flowsheet or bst.main_flowsheet
    funcs = funcs or create_funcs(lactic_tea=lactic_tea, flowsheet=flowsheet)
    s = flowsheet.stream
    u = flowsheet.unit
    flowsheet.system.lactic_sys.simulate()
    u.U101.divert_ratio = 0.25
    MPSP = funcs['simulate_get_MPSP']()
    print('\n---------- Simulation Results ----------')
    print(f'MPSP is ${MPSP:.3f}/kg')
    s.LCA_stream.imass['CH4'] *= 0.75
    s.natural_gas.imass['CH4'] *= 0.75
    GWP = funcs['get_GWP']()
    FEC = funcs['get_FEC']()
    print(f'GWP is {GWP:.3f} kg CO2-eq/kg lactic acid')
    print(f'FEC is {FEC:.2f} MJ/kg lactic acid')
    print('------------------------------------------\n')


def create_funcs(lactic_tea=None, flowsheet=None):
    flowsheet = flowsheet or bst.main_flowsheet
    lactic_sys = flowsheet.system.lactic_sys
    lactic_tea = lactic_tea or create_tea(flowsheet=flowsheet)
    s = flowsheet.stream

    # Simulate system and get results
    def simulate_get_MPSP():
        lactic_sys.simulate()
        return lactic_tea.solve_price(s.lactic_acid)

    funcs = {'simulate_get_MPSP': simulate_get_MPSP}

    ##### LCA #####
    funcs['get_lactic_flow'] = get_lactic_flow = lambda: lactic_sys.get_mass_flow(s.lactic_acid)
    # 100-year global warming potential (GWP) from material flows
    funcs['get_material_GWP'] = get_material_GWP = \
        lambda: lactic_sys.get_total_feeds_impact('GWP')/get_lactic_flow()

    # GWP from onsite emission (e.g., combustion) of non-biogenic carbons
    get_onsite_GWP = lambda: (s.natural_gas.get_atomic_flow('C')+s.ethanol.get_atomic_flow('C')) \
        * 44.0095 / s.lactic_acid.F_mass # 44.0095 is chems.CO2.MW
    funcs['get_onsite_GWP'] = get_onsite_GWP

    # GWP from electricity
    get_electricity_use = lambda: sum(i.power_utility.rate for i in lactic_sys.units) # kW
    funcs['get_electricity_use'] = get_electricity_use
    funcs['get_electricity_GWP'] = get_electricity_GWP = \
        lambda: lactic_sys.get_net_electricity_impact('GWP')/get_lactic_flow()

    funcs['get_GWP'] = lambda: get_material_GWP()+get_onsite_GWP()+get_electricity_GWP()

    # Fossil energy consumption (FEC) from materials
    funcs['get_material_FEC'] = get_material_FEC = \
        lambda: lactic_sys.get_total_feeds_impact('FEC')/get_lactic_flow()

    # FEC from electricity
    funcs['get_electricity_FEC'] = get_electricity_FEC = \
        lambda: lactic_sys.get_net_electricity_impact('FEC')/get_lactic_flow()

    # Total FEC
    get_FEC = lambda: get_material_FEC()+get_electricity_FEC()
    funcs['get_FEC'] = get_FEC

    return funcs


# %%

# def __getattr__(name):
#     if not _chemicals_loaded:
#         _load_chemicals()
#         if name == 'chemicals': return chemicals
#     if not _system_loaded:
#         _load_system()
#         dct = globals()
#         dct.update(flowsheet.to_dict())
#         if name in dct: return dct[name]
#     raise AttributeError(f'module "{__name__}" has no attribute "{name}."')

from .models import *
from . import models

__all__ = (
    'auom', 'CEPCI',
    'flowsheet', 'groups', 'funcs', 'lactic_sys', 'lactic_tea',
    'Area100', 'Area200', 'Area300', 'Area400', 'Area500', 'Area600',
    'simulate_and_print', 'simulate_separation_improvement',
    'simulate_separation_improvement', 'simulate_operating_improvement',
    *_chemicals.__all__,
    *_process_settings.__all__,
    *_tea.__all__,
    *systems.__all__,
    *models.__all__,
    )


# -*- coding: utf-8 -*-
"""

.. contents:: :local:

Fermentation
------------
.. automodule:: biorefineries.cellulosic.systems.fermentation

Pretreatment
------------
.. automodule:: biorefineries.cellulosic.systems.pretreatment

Cellulosic ethanol
------------------
.. automodule:: biorefineries.cellulosic.systems.cellulosic_ethanol

"""
from . import fermentation
from . import pretreatment
from . import cellulosic_ethanol

__all__ = (
    *fermentation.__all__,
    *pretreatment.__all__,
    *cellulosic_ethanol.__all__,
)

from .fermentation import *
from .pretreatment import *
from .cellulosic_ethanol import *



# -*- coding: utf-8 -*-
"""
.. contents:: :local:

Reactors
--------
.. autoclass:: biorefineries.cane.units.SeedTrain
.. autoclass:: biorefineries.cane.units.CoFermentation

Separations
-----------
.. autoclass:: biorefineries.cane.units.OleinCrystallizer
    

"""
import biosteam as bst
from thermosteam import PRxn, Rxn
from biorefineries.cellulosic.units import (
    SeedTrain,
    CoFermentation
)

__all__ = ('SeedTrain', 'CoFermentation', 'OleinCrystallizer')
    
class SeedTrain(SeedTrain):
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, reactions=None, saccharification=False):
        bst.Unit.__init__(self, ID, ins, outs, thermo)
        self.saccharification = saccharification
        chemicals = self.chemicals
        self.reactions = reactions or PRxn([
    #   Reaction definition                   Reactant    Conversion
    Rxn('Glucose -> 2 Ethanol + 2 CO2',       'Glucose',   0.9000, chemicals),
    Rxn('3 Xylose -> 5 Ethanol + 5 CO2',      'Xylose',    0.8000, chemicals),
    Rxn('Glucose -> Cellmass',                'Glucose',  0.0473, chemicals, correct_mass_balance=True),
    Rxn('Xylose -> Cellmass',                 'Xylose',  0.0421, chemicals, correct_mass_balance=True),
        ])
        
    def _setup(self):
        super()._setup()
        self.outs[0].phase = 'g'
        
    def _run(self):
        vent, effluent= self.outs
        effluent.mix_from(self.ins, energy_balance=False)
        self.reactions.force_reaction(effluent)
        effluent.mol.remove_negatives()
        effluent.T = self.T
        vent.empty()
        vent.copy_flow(effluent, ('CO2', 'O2', 'N2'), remove=True)
        

class CoFermentation(CoFermentation):
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 tau=36, N=None, V=3785.4118, T=305.15, P=101325,
                 Nmin=2, Nmax=36, cofermentation=None):
        bst.BatchBioreactor.__init__(self, ID, ins, outs, thermo, tau, N, V, T, P, Nmin, Nmax)
        self.P = P
        chemicals = self.chemicals
        self.loss = None
        self.cofermentation = cofermentation or PRxn([
    #   Reaction definition                   Reactant    Conversion
    Rxn('Glucose -> 2 Ethanol + 2 CO2',       'Glucose',   0.9500, chemicals),
    Rxn('3 Xylose -> 5 Ethanol + 5 CO2',      'Xylose',    0.8500, chemicals),
    Rxn('Glucose -> Cellmass',                'Glucose',  0.05, chemicals, correct_mass_balance=True),
    Rxn('Xylose -> Cellmass',                 'Xylose',  0.05, chemicals, correct_mass_balance=True),
        ])
        
        if 'CSL' in chemicals:
            self.CSL_to_constituents = Rxn(
                'CSL -> 0.5 H2O + 0.25 LacticAcid + 0.25 Protein', 'CSL', 1.0000, chemicals, basis='wt',
            )
            self.CSL_to_constituents.basis = 'mol'
        else:
            self.CSL_to_constituents = None
        if all([i in self.chemicals for i in ('FFA', 'DAG', 'TAG', 'Glycerol')]):
            self.lipid_reaction = self.oil_reaction = PRxn([
                Rxn('TAG + 3Water -> 3FFA + Glycerol', 'TAG', 0.23, chemicals),
                Rxn('TAG + Water -> FFA + DAG', 'TAG', 0.02, chemicals)
            ])
        else:
            self.lipid_reaction = self.oil_reaction = None

          
class AeratedCoFermentation(bst.AeratedBioreactor): # For microbial oil production
    V_max_default = 500
    def __init__(
            self, ID='', ins=None, outs=(), thermo=None,  
            *, cofermentation, theta_O2=0.5, 
            dT_hx_loop=8,
            Q_O2_consumption=-460240, # [kJ/kmol] equivalent to 110 kcal / mol as in https://www.academia.edu/19636928/Bioreactor_Design_for_Chemical_Engineers
            batch=True,
            **kwargs,
        ):
        bst.StirredTankReactor.__init__(self, ID, ins, outs, thermo, batch=batch, dT_hx_loop=dT_hx_loop, **kwargs)
        chemicals = self.chemicals
        self.theta_O2 = theta_O2
        self.hydrolysis_reaction = Rxn('Sucrose + Water -> 2Glucose', 'Sucrose', 1.00, chemicals)
        self.cofermentation = cofermentation
        self.lipid_reaction = self.oil_reaction = PRxn([
            Rxn('TAG + 3Water -> 3FFA + Glycerol', 'TAG', 0.23, chemicals),
            Rxn('TAG + Water -> FFA + DAG', 'TAG', 0.02, chemicals)
        ])
        self.Q_O2_consumption = Q_O2_consumption
        self.optimize_power = False
        self.kLa_coefficients = "Van't Riet"
    
    def _run_vent(self, vent, effluent):
        vent.receive_vent(effluent, energy_balance=False, ideal=True)
    
    def run_reactions(self, effluent):
        self.hydrolysis_reaction.force_reaction(effluent)
        self.lipid_reaction.force_reaction(effluent)
        if effluent.imol['H2O'] < 0.: effluent.imol['H2O'] = 0.
        self.cofermentation.force_reaction(effluent)
        
class AeratedFermentation(bst.AeratedBioreactor): # For microbial oil production
    V_max_default = 500
    def __init__(
            self, ID='', ins=None, outs=(), thermo=None,  
            *, fermentation_reaction, cell_growth_reaction, theta_O2=0.5,
            dT_hx_loop=8,
            Q_O2_consumption=-460240, # [kJ/kmol] equivalent to 110 kcal / mol as in https://www.academia.edu/19636928/Bioreactor_Design_for_Chemical_Engineers
            batch=True,
            **kwargs,
        ):
        bst.StirredTankReactor.__init__(self, ID, ins, outs, thermo, batch=batch, dT_hx_loop=dT_hx_loop, **kwargs)
        chemicals = self.chemicals
        self.theta_O2 = theta_O2
        self.hydrolysis_reaction = Rxn('Sucrose + Water -> 2Glucose', 'Sucrose', 1.00, chemicals)
        self.fermentation_reaction = fermentation_reaction
        self.cell_growth_reaction = cell_growth_reaction
        self.lipid_reaction = self.oil_reaction = PRxn([
            Rxn('TAG + 3Water -> 3FFA + Glycerol', 'TAG', 0.23, chemicals),
            Rxn('TAG + Water -> FFA + DAG', 'TAG', 0.02, chemicals)
        ])
        self.Q_O2_consumption = Q_O2_consumption
        self.optimize_power = False
        self.kLa_coefficients = "Van't Riet"
    
    def _run_vent(self, vent, effluent):
        vent.receive_vent(effluent, energy_balance=False, ideal=True)
    
    def run_reactions(self, effluent):
        self.hydrolysis_reaction.force_reaction(effluent)
        self.lipid_reaction.force_reaction(effluent)
        if effluent.imol['H2O'] < 0.: effluent.imol['H2O'] = 0.
        self.fermentation_reaction.force_reaction(effluent)
        self.cell_growth_reaction.force_reaction(effluent)
    
class OleinCrystallizer(bst.BatchCrystallizer):
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, 
                 T, solid_purity=0.98, melt_purity=0.90,
                 solid_IDs=('TAG', 'FFA', 'PL'), melt_IDs=('AcTAG',),
                 order=None):
        bst.BatchCrystallizer.__init__(self, ID, ins, outs, thermo,
                                       tau=5, V=1e6, T=T)
        self.melt_purity = melt_purity
        self.solid_purity = solid_purity
        self.solid_IDs = solid_IDs
        self.melt_IDs = melt_IDs

    @property
    def Hnet(self):
        feed = self.ins[0]
        effluent = self.outs[0]
        if 's' in feed.phases:
            solid = feed if feed.phase == 's' else feed['s']
            H_in = - sum([i.Hfus * j for i,j in zip(self.chemicals, solid.mol) if i.Hfus])
        else:
            H_in = 0.
        solids = effluent['s']
        H_out = - sum([i.Hfus * j for i,j in zip(self.chemicals, solids.mol) if i.Hfus])
        return H_out - H_in
        
    def _run(self):
        outlet = self.outs[0]
        outlet.phases = ('s', 'l')
        solid_purity = self.solid_purity
        melt_purity = self.melt_purity
        feed = self.ins[0]
        solid_IDs = self.solid_IDs
        melt_IDs = self.melt_IDs
        solid_flows = feed.imass[solid_IDs]
        melt_flows = feed.imass[melt_IDs]
        net_solid_flow = solid_flows.sum()
        net_melt_flow = melt_flows.sum()
        total = net_solid_flow + net_melt_flow
        minimum_melt_purity = net_melt_flow / total
        minimum_solid_purity = net_solid_flow / total
        outlet.empty()
        if solid_purity < minimum_solid_purity:
            outlet.imol['s'] = feed.mol
        elif melt_purity < minimum_melt_purity:
            outlet.imol['l'] = feed.mol
        else: # Lever rule
            solid_purity = (1. - solid_purity)
            melt_fraction = (minimum_melt_purity - solid_purity) / (melt_purity - solid_purity)
            melt = melt_fraction * total
            pure_melt = melt * melt_purity
            impurity_melt = melt - pure_melt
            outlet.imass['l', melt_IDs] = pure_melt * melt_flows / melt_flows.sum()
            outlet.imass['l', solid_IDs] = impurity_melt * solid_flows / solid_flows.sum()
            outlet.imol['s'] = feed.mol - outlet.imol['l']
        outlet.T = self.T
        


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""

import biosteam as bst
from thermosteam import Stream
from biorefineries.cellulosic import chemicals as c
from biorefineries.cellulosic import streams as s
from biorefineries.cellulosic import units
import numpy as np

__all__ = (
    'create_saccharification_system',
)

@bst.SystemFactory(
    ID='saccharification_sys',
    ins=[s.pretreated_biomass,
         s.cellulase,
         s.saccharification_water],
    outs=[s.slurry],
)
def create_saccharification_system(
        ins, outs, solids_loading=None, 
        nonsolids=None,
        insoluble_solids=None,
        insoluble_solids_loading=None,
        ignored=None,
        Saccharification=None,
        saccharification_reactions=None,
    ):
    pretreated_biomass, cellulase, saccharification_water = ins
    slurry, = outs
    if nonsolids is None: nonsolids = c.default_nonsolids
    if insoluble_solids is None: insoluble_solids = c.default_insoluble_solids
    if insoluble_solids_loading is None: insoluble_solids_loading = 10.3
    if solids_loading is None: solids_loading = 0.2
    if ignored is None: ignored = c.default_ignored
    M301 = units.EnzymeHydrolysateMixer('M301', (pretreated_biomass, cellulase, saccharification_water))
    H301 = units.HydrolysateHeatExchanger('H301', M301-0, T=48+273.15)
    M301.solids_loading = solids_loading
    M301.insoluble_solids_loading = insoluble_solids_loading
    M301.enzyme_loading = 0.02 # (20 g enzyme / 1000 g cellulose) 
    M301.enzyme_concentration = 0.05 # (50 g cellulase / 1000g cellulose mixture)
    
    @M301.add_specification
    def update_cellulase_loading():
        pretreated_biomass, cellulase, water = M301.ins
        # Note: An additional 10% is produced to produce sophorose
        # Humbird (2011) pg. 37 
        enzyme_concentration = M301.enzyme_concentration
        enzyme_over_cellulose = M301.enzyme_loading / enzyme_concentration
        z_mass_cellulase_mixture = np.array([1 - enzyme_concentration , enzyme_concentration], float)
        cellulase.imass['Water', 'Cellulase'] = (
            enzyme_over_cellulose
            * z_mass_cellulase_mixture
            * 1.2 * (pretreated_biomass.imass['Glucan'])
        )
    
    
    @M301.add_specification(run=True)
    def update_moisture_content():
        pretreated_biomass, cellulase, saccharification_water, *other = M301.ins
        chemicals = M301.chemicals
        s_mix = Stream.sum([pretreated_biomass, cellulase], None, M301.outs[0].thermo, energy_balance=False)
        mass = s_mix.mol * chemicals.MW
        solids_loading = M301.solids_loading
        insoluble_solids_loading = M301.insoluble_solids_loading
        indices = chemicals.available_indices(nonsolids)
        mass_moisture = mass[indices].sum()
        indices = chemicals.available_indices(insoluble_solids)
        mass_insoluble = mass[indices].sum()
        indices = chemicals.available_indices(ignored)
        if indices:
            mass_ignored = mass[indices].sum()
        else:
            mass_ignored = 0.
        total_mass = mass.sum() - mass_ignored
        water_over_solids = (1 - solids_loading) / solids_loading
        water_over_insoluble_solids = (1 - insoluble_solids_loading) / insoluble_solids_loading
        M301.required_saccharification_water = max(
            water_over_solids * (total_mass - mass_moisture) - mass_moisture,
            water_over_insoluble_solids * mass_insoluble - (total_mass - mass_insoluble)
        )
        saccharification_water.imass['Water'] = max(M301.required_saccharification_water, 0.)
    
    if Saccharification is None:
        Saccharification = units.ContinuousPresaccharification
    Saccharification('R301', H301-0, slurry, reactions=saccharification_reactions)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun  4 23:44:10 2021

@author: yrc2
"""
from biorefineries import cane
import biosteam as bst
import numpy as np
from biosteam.utils import colors, GG_colors
import matplotlib.pyplot as plt
from biosteam.plots import (
    rounded_linspace,
    generate_contour_data,
    plot_contour_2d,
    MetricBar,
    plot_vertical_line,
    rounded_tickmarks_from_data as tickmarks,
    plot_scatter_points,
    plot_contour_single_metric,
)
from thermosteam.units_of_measure import format_units
from thermosteam.utils import set_figure_size, set_font
from .results import images_folder
from . import feature_mockups as feature
from .data import microbial_oil_baseline as perf
from warnings import filterwarnings
from scipy.ndimage.filters import gaussian_filter
from colorpalette import ColorWheel
import os

__all__ = (
    'plot_sorghum_oil_content_and_cane_oil_content_contours_manuscript',
    'plot_metrics_across_composition_manuscript',
    'plot_sorghum_oil_content_and_cane_oil_content_contours_seminar',
    'plot_recovery_and_oil_content_contours_manuscript',
    'plot_recovery_and_oil_content_contours',
    'plot_sorghum_oil_content_and_cane_oil_content_contours',
    'plot_recovery_and_oil_content_contours_biodiesel_only',
    'plot_recovery_and_oil_content_contours_with_oilsorghum_only',
    'plot_metrics_across_biomass_yield',
    'plot_metrics_across_composition',
    'plot_oil_recovery_integration',
)

filterwarnings('ignore', category=bst.exceptions.DesignWarning)
line_color_wheel = ColorWheel([
    GG_colors.orange,
    GG_colors.purple,
    GG_colors.green,
    GG_colors.blue,
    GG_colors.yellow,
    colors.CABBI_teal,
    colors.CABBI_grey,
    colors.CABBI_brown,
])
for i, color in enumerate(line_color_wheel.colors): line_color_wheel.colors[i] = color.tint(20)
shadecolor = (*colors.neutral.RGBn, 0.20)
linecolor = (*colors.neutral_shade.RGBn, 0.85)
targetcolor = (*colors.red_tint.RGBn, 1)
startcolor = (*colors.red_tint.RGBn, 1)
edgecolor = (*colors.CABBI_black.RGBn, 1)
    
CABBI_colors = (colors.CABBI_yellow.tint(75).RGBn, 
                colors.CABBI_yellow.tint(30).RGBn,
                colors.CABBI_yellow.RGBn,
                colors.CABBI_green.tint(5).RGBn,
                colors.CABBI_teal_green.shade(40).RGBn,
                colors.CABBI_teal_green.shade(60).RGBn)

CABBI_colors_x = (colors.CABBI_blue_light.tint(90).RGBn,
                  colors.CABBI_blue_light.tint(40).RGBn, 
                  colors.CABBI_blue_light.RGBn, 
#                   colors.CABBI_teal.RGBn,
#                   colors.CABBI_teal_green.tint(10).RGBn,
                  colors.CABBI_teal_green.tint(40).shade(15).RGBn,
                  colors.CABBI_teal_green.shade(45).RGBn)

diverging_colormaps = [
    plt.cm.get_cmap('RdYlGn')
]

colormaps = [
    plt.cm.get_cmap('viridis'),
    plt.cm.get_cmap('copper_r'),
    # LinearSegmentedColormap.from_list('CABBI', CABBI_colors, 25),
    # LinearSegmentedColormap.from_list('CABBI', CABBI_colors_x, 25),
    plt.cm.get_cmap('inferno_r'),
    plt.cm.get_cmap('copper_r'),
    plt.cm.get_cmap('bone_r'),
] * 2

light_letter_color = colors.neutral.tint(98).RGBn
letter_color = colors.neutral.tint(80).RGBn
dark_letter_color = colors.neutral.shade(80).RGBn

default_box_color = colors.neutral.RGBn

# %% Plot functions for publication

def _add_letter_labels(axes, xpos, ypos, colors, box_color=None):
    M, N = shape = colors.shape
    if shape == (2, 2):
        letters=np.array([['A', 'C'], ['B', 'D']])
    elif shape == (3, 2):
        letters=np.array([['A', 'D'], ['B', 'E'], ['C', 'F']])
    elif shape == (2, 3):
        letters=np.array([['A', 'C', 'E'], ['B', 'D', 'F']])
    elif shape == (2, 1):
        letters=np.array([['A'], ['B']])
    if box_color is None:
        box_color = default_box_color
    txtbox = dict(boxstyle='round', facecolor=box_color, 
                  edgecolor='None', alpha=0.6)
    letters = 'ABCDEFG'
    n = -1
    for i in range(M):
        for j in range(N):
            n += 1
            letter = letters[n]
            ax = axes[i, j]
            xlb, xub = ax.get_xlim()
            ylb, yub = ax.get_ylim()
            if hasattr(ax, '_cached_ytwin'):                
                ax = ax._cached_ytwin
            ax.text((xlb + xub) * xpos, (yub + ylb) * ypos, letter, color=colors[i, j],
                     horizontalalignment='center',verticalalignment='center',
                     fontsize=12, fontweight='bold', zorder=1, bbox=txtbox)

def plot_metrics_across_composition_manuscript(load=True, fs=8, smooth=1):
    set_font(size=fs)
    set_figure_size()
    fig, axes = plot_metrics_across_composition(
        load=load, 
        smooth=smooth,
    )
    colors = np.zeros([2, 2], object)
    colors[:] = [[light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color]]
    _add_letter_labels(axes, 1 - 0.86, 0.61, colors)
    plt.subplots_adjust(right=0.92, hspace=0.1, wspace=0.1, top=0.9, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'metrics_across_composition_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_recovery_and_oil_content_contours_manuscript(load=True, fs=8, smooth=1):
    set_font(size=fs)
    set_figure_size()
    fig, axes = plot_recovery_and_oil_content_contours(
        load=load, 
        smooth=smooth,
    )
    colors = np.zeros([2, 2], object)
    colors[:] = [[light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color]]
    _add_letter_labels(axes, 1 - 0.68, 0.7, colors)
    plt.subplots_adjust(right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'recovery_and_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_recovery_and_oil_content_contours_with_oilsorghum_only(fs=10, smooth=1):
    set_font(size=fs)
    set_figure_size(4, 1.1)
    fig, axes = plot_recovery_and_oil_content_contours(
        load=True, 
        smooth=smooth,
        with_oilsorghum_only=True
    )
    colors = np.zeros([2, 1], object)
    colors[:] = [[light_letter_color], [light_letter_color]]
    # _add_letter_labels(axes, 1 - 0.68, 0.7, colors)
    plt.subplots_adjust(left=.2, right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'recovery_and_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_recovery_and_oil_content_contours_biodiesel_only(load=True, fs=8, metric_indices=None):
    set_font(size=fs)
    set_figure_size()
    if metric_indices is None: metric_indices = (0, 2, 6, 10)
    for cmap, i in zip(colormaps, metric_indices):
        fig, axes = plot_recovery_and_oil_content_contours(
            load=load, configurations=np.array([[7, 8], [5, 6]]),
            N_points=20, yticks=[0, 2.5, 5, 7.5, 10, 12.5, 15],
            titles=['Batch', 'Fed-Batch'],
            metric_index=i, cmap=cmap,
        )
        load = True
        colors = np.zeros([2, 2], object)
        colors[:] = [[light_letter_color, light_letter_color],
                     [light_letter_color, light_letter_color]]
        _add_letter_labels(axes, 1 - 0.68, 0.85, colors)
        plt.subplots_adjust(right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
        for j in ('svg', 'png'):
            file = os.path.join(images_folder, f'recovery_and_oil_content_contours_biodiesel_only_{i}.{j}')
            plt.savefig(file, transparent=True)

def plot_sorghum_oil_content_and_cane_oil_content_contours_manuscript(load=True, fs=8, smooth=0.9, configuration_index=None):
    set_font(size=fs)
    set_figure_size()
    fig, axes = plot_sorghum_oil_content_and_cane_oil_content_contours(
        load=load, smooth=smooth, configuration_index=configuration_index,
    )
    colors = np.zeros([2, 2], object)
    colors[:] = [[light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color]]
    _add_letter_labels(axes, 1 - 0.75, 0.75, colors)
    plt.subplots_adjust(left=0.12, right=0.92, wspace=0.12, hspace=0.12, top=0.9, bottom=0.12)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'relative_sorghum_oil_content_and_cane_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_sorghum_oil_content_and_cane_oil_content_contours_seminar(load=True):
    plot_sorghum_oil_content_and_cane_oil_content_contours_manuscript(load, 10, 1.3, [2, 6])

# %% General    

def plot_metrics_across_composition(
        load=False, N_decimals=1, 
        yticks=None, titles=None, 
        cmap=None, smooth=None,
    ):
    xticks = [0,   2,   4,   6,   8,   10]
    xlim = (xticks[0], xticks[-1])
    yticks = [35,  45,  55,  65,  75]
    ylim = (yticks[0], yticks[-1])
    metrics = COBY, NEP = [cane.competitive_biomass_yield, cane.net_energy_production]
    metric_indices = [cane.all_metric_mockups.index(i) for i in metrics] 
    X, Y, data = generate_contour_data(
        cane.evaluate_metrics_at_composition, 
        xlim, ylim, load=load, n = 10,
    )
    data = data[:, :, :, metric_indices]
    data[:, 0, 0, 0] = data[:, 1, 0, 0] + (data[:, 1, 0, 0] - data[:, 2, 0, 0])
    data[:, 0, 1, 0] = data[:, 1, 1, 0] + (data[:, 1, 1, 0] - data[:, 2, 1, 0])
    
    # Plot contours
    xlabel = 'Oil content [dry wt. %]'
    ylabel = "Fiber content [dry wt. %]"
    if titles is None: titles = np.array(['Bioethanol', 'Microbial oil'])
    d0 = data[:, :, 0, :]
    d1 = data[:, :, 1, :]
    metric_bars = [
        # MetricBar(MFPP.name, format_units(MFPP.units), colormaps[1], tickmarks(data[:, :, 0, :], 5, 1, expand=0, p=0.5), 18, 1),
        MetricBar('Comp. biomass yield', 'Dry-'+format_units('MT/ha'), plt.cm.get_cmap('viridis_r'), tickmarks(d0[~np.isnan(d0)], 5, 1, expand=0, p=0.5), 10, 1),
        # MetricBar('Biod. prod.', format_units(BP.units), plt.cm.get_cmap('copper'), tickmarks(data[:, :, 1, :], 8, 1, expand=0, p=0.5), 10, 1),
        MetricBar('Net energy prod.', format_units(NEP.units), plt.cm.get_cmap('inferno'), tickmarks(d1[~np.isnan(d1)], 5, 1, expand=0, p=2), 20, 1),
    ]
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X, 100.*Y, titles, data, xlabel, ylabel, xticks, yticks, metric_bars, 
        styleaxiskw=dict(xtick0=True), label=True, wbar=2.1
    )
    try:
        df = cane.get_composition_data()
    except:
        pass
    else:
        names = df.index
        lines = []
        for name, color in zip(names, line_color_wheel):
            data = df.loc[name]
            if str(name) in ('19B', '316'): continue
            oil = data['Stem oil (dw)']['Mean'] * 100
            lines.append(
                (name, 
                 oil, 
                 data['Fiber (dw)']['Mean'] * 100,
                 data['Water (wt)']['Mean'] * 100,
                 color.RGBn)
            )
        for i in axes.flat: 
            if hasattr(i, '_cached_ytwin'):
                plt.sca(i); plt.xlim(xlim)
                plt.sca(i._cached_ytwin); plt.xlim(xlim)
        txtbox = dict(boxstyle='round', facecolor=colors.neutral.shade(20).RGBn, 
                      edgecolor='None', alpha=0.999, pad=0.2)
        
        for *axes_columns, _ in axes:
            for (name, lipid, fiber, moisture, color) in lines:
                left, right = axes_columns
                plt.sca(right._cached_ytwin)
                plt.text(
                    lipid + 0.1, fiber + 1, name, weight='bold', c=color,
                    bbox=txtbox,
                )
                plot_scatter_points(
                    [lipid], [fiber], marker='X', s=50, 
                    color=color, edgecolor=edgecolor, clip_on=False, zorder=1e6,
                )
                
                plt.sca(left)
                plt.xlim([0.5, 10])
                if lipid > 0.5:
                    plt.text(
                        lipid + 0.1, fiber + 1, name, weight='bold', c=color,
                        bbox=txtbox,
                    )
                    plot_scatter_points(
                        [lipid], [fiber], marker='X', s=50, 
                        color=color, edgecolor=edgecolor, clip_on=False, zorder=1e6,
                    )
                    
    return fig, axes

def plot_metrics_across_biomass_yield(
        load=False, N_decimals=1, 
        yticks=None, titles=None, 
        cmap=None, smooth=None,
    ):
    xticks = [0,   2,   4,   6,   8,   10]
    xlim = (xticks[0], xticks[-1])
    yticks = [5, 10, 15, 20, 25, 30]
    ylim = (yticks[0], yticks[-1])
    metrics = MBSP, = [cane.MBSP]
    metric_indices = [cane.all_metric_mockups.index(i) for i in metrics] 
    X, Y, data = generate_contour_data(
        cane.evaluate_metrics_at_biomass_yield,
        xlim, ylim, file=contour_file('configurations_biomass_yield_analysis'),
        smooth=smooth,
    )
    data = data[:, :, metric_indices, :]
    data = np.swapaxes(data, 2, 3)
    # Plot contours
    xlabel = 'Oil content [dry wt. %]'
    ylabel = "Biomass yield [dry wt. %]"
    if titles is None: titles = np.array(['Direct Cogeneration', 'Integrated Cofermentation'])
    d0 = data[:, :, 0, :]
    metric_bars = [
        # MetricBar(MFPP.name, format_units(MFPP.units), colormaps[1], tickmarks(data[:, :, 0, :], 5, 1, expand=0, p=0.5), 18, 1),
        MetricBar('MBSP', format_units('USD/L'), plt.cm.get_cmap('viridis_r'), tickmarks(d0[~np.isnan(d0)], 5, 1, expand=0, p=0.5), 15, 1),
    ]
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X, Y, titles, data, xlabel, ylabel, xticks, yticks, metric_bars, 
        styleaxiskw=dict(xtick0=True), label=True, wbar=2.1
    )
    try:
        df = cane.get_composition_data()
    except:
        pass
    else:
        names = df.index
        lines = []
        for name, color in zip(names, line_color_wheel):
            data = df.loc[name]
            if str(name) in ('19B', '316'): continue
            oil = data['Stem oil (dw)']['Mean'] * 100
            lines.append(
                (name, 
                 oil, 
                 data['Biomass yield (dry MT/ha)']['Mean'],
                 color.RGBn)
            )
        for i in axes.flat: 
            if hasattr(i, '_cached_ytwin'):
                plt.sca(i); plt.xlim(xlim)
                plt.sca(i._cached_ytwin); plt.xlim(xlim)
        txtbox = dict(boxstyle='round', facecolor=colors.neutral.shade(20).RGBn, 
                      edgecolor='None', alpha=0.999, pad=0.2)
        
        for *axes_columns, _ in axes:
            for (name, lipid, biomass_yield, color) in lines:
                left, right = axes_columns
                plt.sca(right._cached_ytwin)
                plt.text(
                    lipid + 0.1, biomass_yield + 1, name, weight='bold', c=color,
                    bbox=txtbox,
                )
                plot_scatter_points(
                    [lipid], [biomass_yield], marker='X', s=50, 
                    color=color, edgecolor=edgecolor, clip_on=False, zorder=1e6,
                )
                
                plt.sca(left._cached_ytwin)
                plt.text(
                    lipid + 0.1, biomass_yield + 1, name, weight='bold', c=color,
                    bbox=txtbox,
                )
                plot_scatter_points(
                    [lipid], [biomass_yield], marker='X', s=50, 
                    color=color, edgecolor=edgecolor, clip_on=False, zorder=1e6,
                )
                    
    return fig, axes

def relative_sorghum_oil_content_and_cane_oil_content_data(load, configurations):
    # Generate contour data
    y = np.linspace(0.01, 0.10, 20)
    x = np.linspace(0.01, 0.10, 20)
    X, Y = np.meshgrid(x, y)
    folder = os.path.dirname(__file__)
    config = ''.join([str(i) for i in configurations])
    file = f'oil_content_analysis_{config}.npy'
    file = os.path.join(folder, file)
    if configurations is None: configurations = [1, 2]
    if load:
        data = np.load(file)
    else:
        from warnings import filterwarnings; filterwarnings('ignore')
        data = cane.evaluate_configurations_across_sorghum_and_cane_oil_content(
            X, Y, configurations, 
        )
        np.save(file, data)
    return X, Y, data
    
def contour_file(name):
    folder = os.path.dirname(__file__)
    file = name + '.npy'
    return os.path.join(folder, file)

def plot_fermentation_performance(
        load=False, 
    ):
    metrics = [feature.MBSP, feature.GWP_biodiesel]
    metrics_index = [feature.all_metric_mockups.index(i) for i in metrics]
    X, Y, Z = generate_contour_data(
        cane.evaluate_fermentation_performance, # TODO
        xlim=[50, 90],
        ylim=[100 * perf.min_lipid_yield_glucose, 100 * perf.max_lipid_yield_glucose],
        file=contour_file('oil_recovery_integration'),
        n=10,
        load=load,
    )
    Z = Z[:, :, metrics_index, :]
    # Plot contours
    xlabel = 'Microbial oil recovery [%]'
    ylabel = "Microbial oil yield [wt. %]"
    xticks = [50, 60, 70, 80, 90]
    yticks = rounded_linspace(
        100 * perf.min_lipid_yield_glucose, 100 * perf.max_lipid_yield_glucose, 5, 
    )
    titles = ['Mechanical oil recovery', 'Integrated oil recovery']
    
    d0 = Z[:, :, 0, :]
    d1 = Z[:, :, 1, :]
    d2 = Z[:, :, 2, :]
    metric_bars = [
        MetricBar('MBSP', format_units('USD/L'), plt.cm.get_cmap('viridis_r'), tickmarks(d0[~np.isnan(d0)], 5, 1, expand=0, p=0.5), 15, 1),
        MetricBar('ROI', format_units('% USD'), plt.cm.get_cmap('viridis'), tickmarks(d1[~np.isnan(d1)], 5, 1, expand=0, p=1), 15, 1),
        MetricBar("GWP", format_units('USD/L'), plt.cm.get_cmap('inferno_r'), tickmarks(d2[~np.isnan(d2)], 5, 0.1, expand=0, p=0.1), 15, 1),
    ]
    fig, axes, CSs, CB, other_axes = plot_contour_2d(
        X, Y, Z, xlabel, ylabel, xticks, yticks, metric_bars,  titles,
        fillcolor=None, styleaxiskw=dict(xtick0=False), label=True,
    )
    return fig, other_axes

def plot_oil_recovery_integration(
        load=False, 
    ):
    productivity = np.array([
        perf.hydrolysate_productivity, 
        perf.batch_productivity_mean, 
        perf.fed_batch_productivity_mean
    ])
    X, Y, Z = generate_contour_data(
        cane.evaluate_metrics_oil_recovery_integration,
        xlim=[50, 90],
        ylim=[100 * perf.min_lipid_yield_glucose, 100 * perf.max_lipid_yield_glucose],
        args=(
            productivity,
        ),
        file=contour_file('oil_recovery_integration'),
        load=load,
        n=10,
    )
    metric = feature.biodiesel_production
    metric_index = cane.all_metric_mockups.index(metric)
    Z = Z[..., metric_index]
    # Plot contours
    xlabel = 'Microbial oil recovery [%]'
    ylabel = "Microb. oil yield [wt. %]"
    units = r'$g \cdot L^{-1} \cdot h^{-1}$'
    ylabels = [f"{ylabel}\nat {round(i, 2)} {units}" for i in productivity]
    xticks = [50, 60, 70, 80, 90]
    yticks = rounded_linspace(
        100 * perf.min_lipid_yield_glucose, 100 * perf.max_lipid_yield_glucose, 5, 
    )
    titles = ['Mechanical oil recovery', 'Integrated oil recovery']
    metric_bar = MetricBar(
        metric.name, format_units(metric.units), plt.cm.get_cmap('viridis_r'), tickmarks(Z[~np.isnan(Z)], 5, 1, expand=0, p=0.5), 15, 1
    )
    fig, axes, CSs, CB, other_axes = plot_contour_single_metric(
        X, Y, Z, xlabel, ylabels, xticks, yticks, metric_bar,  titles,
        fillcolor=None, styleaxiskw=dict(xtick0=False), label=True,
    )
    return fig, other_axes

def plot_sorghum_oil_content_and_cane_oil_content_contours(
        load=False, configuration_index=None, smooth=None,
    ):
    if configuration_index is None: configuration_index = [1, 2]
    # Generate contour data
    X, Y, data = relative_sorghum_oil_content_and_cane_oil_content_data(load, configuration_index)
    MFPP = cane.MFPP
    TCI = cane.TCI
    metrics = [MFPP, TCI]
    metric_indices = [cane.all_metric_mockups.index(i) for i in metrics]
    data = data[:, :, :, metric_indices]
    
    # Plot contours
    xlabel = "Oil-sorghum oil content [dry wt. %]" 
    ylabel = 'Oilcane oil content\n[dry wt. %]'
    yticks = [2, 4, 6, 8, 10]
    xticks = [2, 4, 6, 8, 10]
    if configuration_index == [1, 2]:
        Z = np.array(["Direct Cogeneration", "Integrated Co-Fermentation"])
        data = np.swapaxes(data, 2, 3)
    elif configuration_index == [2, 6]:
        Z = np.array(["Bioethanol", "Microbial oil"])
        data = np.swapaxes(data, 2, 3)
    else:
        raise ValueError('configuration index must be either [1, 2] or [2, 6]')
    
    if smooth: # Smooth curves due to heat exchanger network and discontinuities in design decisionss
        A, B, M, N = data.shape
        for m in range(M):
            for n in range(N):
                metric_data = data[:, :, m, n]
                if isinstance(smooth, int):
                    for a in range(A):
                        values = metric_data[a, :]
                        values.sort()
                        p = np.arange(values.size)
                        coeff = np.polyfit(p, values, smooth)
                        values[:] = np.polyval(coeff, p)
                    for b in range(B):
                        values = metric_data[:, b]
                        values.sort()
                        p = np.arange(values.size)
                        coeff = np.polyfit(p, values, smooth)
                        values[:] = np.polyval(coeff, p)
                else:
                    data[:, :, m, n] = gaussian_filter(metric_data, smooth)
                    
    metric_bars = [
        # [MetricBar(MFPP.name, format_units(MFPP.units), colormaps[0], tickmarks(data[:, :, 0, 0], 5, 1, expand=0, p=0.5), 10, 1),
         MetricBar(MFPP.name, format_units(MFPP.units), colormaps[0], tickmarks(data[:, :, 0, [0, 1]], 5, 1, expand=0, p=0.5), 12, 1),
        # [MetricBar(TCI.name, format_units(TCI.units), colormaps[1], tickmarks(data[:, :, 1, 0], 5, 5, expand=0, p=5), 10, 1),
         MetricBar(TCI.name, format_units(TCI.units), colormaps[1], tickmarks(data[:, :, 1, [0, 1]], 5, 5, expand=0, p=5), 15, 1),
    ]
    
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X, 100.*Y, Z, data, xlabel, ylabel, xticks, yticks, metric_bars, 
        styleaxiskw=dict(xtick0=True), label=True,
    )
    for *cols, _ in axes:
        for i in cols:
            plt.sca(i)
            plt.xlim([1, 10])
    # for i in axes.flatten():
    #     plt.sca(i)
    #     plot_scatter_points([7], [10], marker='*', s=100, color=startcolor,
    #                         edgecolor=edgecolor, clip_on=False, zorder=3)
    return fig, axes
    
def plot_recovery_and_oil_content_contours(
        load=False, metric_index=0, N_decimals=1, configurations=None,
        N_points=20, yticks=None, titles=None, cmap=None, smooth=None,
        with_oilsorghum_only=False,
    ):
    if yticks is None: yticks = [5, 7.5, 10, 12.5, 15]
    if configurations is None:
        if with_oilsorghum_only:
            configurations = np.array([['O1*'], ['O2*']])
            if titles is None: titles = ['Oilcane & Oil-sorghum']
        else:
            configurations = np.array([['O1', 'O1*'], ['O2', 'O2*']])
            if titles is None: titles = ['Oilcane Only', 'Oilcane & Oil-sorghum']
        
    # Generate contour data
    x = np.linspace(0.40, 1.0, N_points)
    y = np.linspace(yticks[0] / 100, yticks[-1] / 100, N_points)
    X, Y = np.meshgrid(x, y)
    metric = bst.metric
    folder = os.path.dirname(__file__)
    file = "oil_extraction_analysis.npy"
    file = os.path.join(folder, file)
    
    if load:
        data = np.load(file)
    else:
        data = cane.evaluate_configurations_across_recovery_and_oil_content(
            X, Y, configurations,
        )
    np.save(file, data)
    data = data[:, :, :, :, metric_index]
    
    if with_oilsorghum_only:
        data = data[:, :, :, -1:]
    
    if smooth: # Smooth curves due to heat exchanger network and discontinuities in design decisionss
        A, B, M, N = data.shape
        for m in range(M):
            for n in range(N):
                metric_data = data[:, :, m, n]
                data[:, :, m, n] = gaussian_filter(metric_data, smooth)
    
    # Plot contours
    xlabel = 'Crushing mill oil recovery [%]'
    ylabel = "Oil content [dry wt. %]"
    ylabels = [f'Direct Cogeneration\n{ylabel}',
               f'Integrated Co-Fermentation\n{ylabel}']
    xticks = [40, 50, 60, 70, 80, 90, 100]
    
    metric = cane.all_metric_mockups[metric_index]
    units = metric.units if metric.units == '%' else format_units(metric.units)
    if cmap is None: cmap = colormaps[metric_index]
    mb = lambda x, name=None: MetricBar(
        metric.name if name is None else name, units if name is None else "", cmap, 
        tickmarks(data[:, :, x, :], 5, 0.1, expand=0, p=0.1, f=lambda x: round(x, N_decimals)),
        10, N_decimals=N_decimals
    )
    
    metric_bars = [mb(0), mb(1, "")]
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X, 100.*Y, titles, data, xlabel, ylabels, xticks, yticks, metric_bars, 
        fillcolor=None, styleaxiskw=dict(xtick0=False), label=True,
    )
    M, N = configurations.shape
    for i in range(M):
        for j in range(N):
            ax = axes[i, j]
            plt.sca(ax)
            plt.fill_between([60, 90], [yticks[0]], [yticks[-1]], 
                              color=shadecolor,
                              linewidth=1)
            plot_vertical_line(60, ls='-.',
                                color=linecolor,
                                linewidth=1.0)
            plot_vertical_line(90, ls='-.',
                               color=linecolor,
                               linewidth=1.0)
            if hasattr(ax, '_cached_ytwin'):                
                plt.sca(ax._cached_ytwin)
            # plot_scatter_points([60], [10], marker='*', s=100, color=startcolor,
            #                     edgecolor=edgecolor, clip_on=False, zorder=3)
            # plot_scatter_points([ub], [15], marker='*', s=100, color=targetcolor,
            #                     edgecolor=edgecolor, clip_on=False, zorder=3)
    return fig, axes


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
#                     Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from .chemicals import create_cellulosic_ethanol_chemicals
from .process_settings import load_process_settings
from .systems import create_cellulosic_ethanol_system
from biorefineries.tea import create_cellulosic_ethanol_tea
from biosteam import main_flowsheet as F

__all__ = (
    'Biorefinery',
)

ethanol_density_kgL = 0.789 # kg/L
liter_per_gallon = 3.78541
ethanol_density_kggal = liter_per_gallon * ethanol_density_kgL # kg/gal

class Biorefinery:
    cache = {}
    
    @property
    def chemicals(self):
        try:
            chemicals = self._chemicals
        except AttributeError:
            self._chemicals = chemicals = create_cellulosic_ethanol_chemicals()
        return chemicals
    
    def __new__(cls, name=None, cache=cache, chemicals=None, include_blowdown_recycle=None,
                feedstock_kwargs=None, prices=None, GWP_CFs=None):
        if include_blowdown_recycle is None: include_blowdown_recycle = False
        if name is None: name = 'corn stover ethanol'
        if 'ethanol' not in name:
            raise ValueError(f"configuration '{name}' is not available; "
                              "only 'ethanol' is valid")
        
        key = (name, include_blowdown_recycle)
        if key in cache:
            return cache[key]
        else:
            cache[key] = self = super().__new__(cls)
        if chemicals is not None: self._chemicals = chemicals
        self.flowsheet = bst.Flowsheet(name)
        F.set_flowsheet(self.flowsheet)
        bst.settings.set_thermo(self.chemicals)
        load_process_settings()
        sys = self.sys = create_cellulosic_ethanol_system(
            include_blowdown_recycle=include_blowdown_recycle
        )
        # Replace feedstock
        stream = F.stream
        if feedstock_kwargs:
            cornstover = stream.search('cornstover')
            feedstock_kwargs['ID'] = feedstock_kwargs.get('ID') or 'feedstock'
            feedstock_kwargs['total_flow'] = feedstock_kwargs.get('total_flow') or cornstover.F_mass
            feedstock = bst.Stream(**feedstock_kwargs)
            sink = cornstover.sink
            cornstover.disconnect_sink()
            sink.ins[0] = feedstock
            stream.discard(cornstover)
        else:
            self.cornstover_sys = sys # Feedstock defaults to corn stover
        # Update TEA and LCA settings
        if prices:
            prices = prices.copy()
            e_price = prices.pop('Electricity', None)
            if e_price: bst.PowerUtility.price = e_price
            for ID, price in prices.items(): stream.search(ID).price = price
        if GWP_CFs:
            GWP_CFs = GWP_CFs.copy()
            e_CF = GWP_CFs.pop('Electricity', None)
            if e_CF: bst.PowerUtility.characterization_factors['GWP'] = e_CF
            for ID, CF in GWP_CFs.items(): stream.search(ID).characterization_factors['GWP'] = CF
        sys.simulate()
        u = F.unit
        OSBL_units = (u.WWTC, u.CWP, u.CT, u.PWC, u.ADP,
                      u.T701, u.T702, u.P701, u.P702, u.M701, u.FWT,
                      u.CSL_storage, u.DAP_storage, u.BT)
        self.tea = self.cornstover_tea = tea = create_cellulosic_ethanol_tea(
            sys, OSBL_units=OSBL_units
        )
        ethanol = F.stream.ethanol
        ethanol.price = tea.solve_price(ethanol)
        self.ethanol_price_gal = ethanol.price * ethanol_density_kggal
        UnitGroup = bst.process_tools.UnitGroup
        self.Area100 = UnitGroup('Area 100', (u.U101,))
        self.Area200 = UnitGroup('Area 200', (u.T201, u.M201, u.R201, u.P201, u.P202,
                                              u.T202, u.F201, u.H201, u.T203, u.M205))
        self.Area300 = UnitGroup('Area 300', (u.H301, u.M301, u.R301,
                                              u.R302, u.R303, u.T301, u.T302))                 
        self.Area400 = UnitGroup('Area 400', (u.D401, u.H401, u.D402, u.P401,
                                              u.M402, u.D403, u.P402, u.H402,
                                              u.U401, u.H403, u.M701, u.S401,
                                              u.P403))
        self.Area500 = UnitGroup('Area 500', (u.WWTC,))
        self.Area600 = UnitGroup('Area 600', (u.T701, u.T702, u.P701, u.P702, u.M701, u.FWT,
                                              u.CSL_storage, u.DAP_storage, u.T703,
                                              u.Ammonia_storage, u.H2SO4_storage))
        self.Area700 = UnitGroup('Area 700', (u.BT,))
        self.Area800 = UnitGroup('Area 800', (u.CWP, u.CT, u.PWC, u.ADP, u.CIP))
        self.areas = (self.Area100, self.Area200, self.Area300, self.Area400,
                      self.Area500, self.Area600, self.Area700, self.Area800)
        WWTsys = sys.find_system(u.R602)
        WWTsys.set_tolerance(method='fixed-point', maxiter=1000, mol=10)
        self.__dict__.update(self.flowsheet.to_dict())
        return self

    


# -*- coding: utf-8 -*-
"""
Created on Thu Mar 30 18:08:06 2023

@author: yrc2
"""
### Data from https://link.springer.com/article/10.1007/s00253-016-7815-y#Sec2
gL_OD600 = 0.64 # From correspondance with authors [g / L / OD600]

### Seed
# Note: assume no lipid produced in seed, only biomass
seed_volume = 25 # mL
seed_concentration = 1 # OD600
seed_amount = seed_volume * seed_concentration * gL_OD600 # 16 g  
seed_time = 48 # hr

# Note: this doesn't have to be the case industrially (and for the model)
# feed/substrate to fermentation can be diverted to seed train.
seed_glucose = 1.801 # g / L

### Batch fermentation
batch_volume = 400 # mL
oxygen_suplied = True # Maintained at 0.50 dissolved oxygen
seed_supplied = seed_amount / (batch_volume + seed_volume) # 0.03765 g / mL

batch_glucose = 150 # g / L
batch_titer_mean = 27.4 # g Lipid / L
batch_titer_std = 0.7 # g Lipid / L
batch_lipid_content_mean = 0.511 # g Lipid / g microbes
batch_lipid_content_std = 0.0026 # g Lipid / g microbes
batch_productivity_mean = 0.31 # g Lipid / L / hr
batch_productivity_std = 0.01 # g Lipid / L / hr
batch_lipid_yield_mean = 0.18 # g / g 
batch_lipid_yield_std = 0.01 # g / g 

fed_batch_glucose = (60 + 80 * 4) # g / L
fed_batch_titer_mean = 89.4 # g Lipid / L
fed_batch_titer_std = 4.5 # g Lipid / L
fed_batch_lipid_content_mean = 0.756 # g Lipid / g microbes
fed_batch_lipid_content_std = 0.0039 # g Lipid / g microbes
fed_batch_productivity_mean = 0.62 # g Lipid / L / hr
fed_batch_productivity_std = 0.03 # g Lipid / L / hr
fed_batch_lipid_yield_mean = 0.22 # g / g 
fed_batch_lipid_yield_std = 0.01 # g / g 

### Computed values for simulation:
feed_diverted_to_seed = seed_volume / (batch_volume + seed_volume) # 0.0588 by vol

### Useful values not used in simulation:
batch_biomass_yield_mean = batch_lipid_yield_mean / batch_lipid_content_mean # 0.357 g microbe / g glucose
batch_CO2_yield_mean = 1 - batch_biomass_yield_mean - batch_lipid_yield_mean # 0.459 g CO2 / g glucose
batch_biomass_growth_coefficient_mean = batch_biomass_yield_mean / (batch_biomass_yield_mean + batch_CO2_yield_mean) # 0.437 g Biomass / g CO2-Biomass
fed_batch_biomass_yield_mean = fed_batch_lipid_yield_mean / fed_batch_lipid_content_mean # 0.357 g microbe / g glucose
fed_batch_CO2_yield_mean = 1 - fed_batch_biomass_yield_mean - fed_batch_lipid_yield_mean # 0.459 g CO2 / g glucose
fed_batch_biomass_growth_coefficient_mean = fed_batch_biomass_yield_mean / (fed_batch_biomass_yield_mean + fed_batch_CO2_yield_mean) # 0.437 g CO2 / g CO2-Biomass

### Data for hydrolysate
hydrolysate_titer = 13.2 # g Lipid / L
hydrolysate_lipid_content = 0.385 # by wt
hydrolysate_sugar_concentration = 100 # g / L
hydrolysate_time = 77.5 # hr
hydrolysate_productivity = hydrolysate_titer / hydrolysate_time # 0.17 g / L / hr
hydrolysate_lipid_yield = hydrolysate_titer / hydrolysate_sugar_concentration # 0.132 g / g
hydrolysate_biomass_yield = hydrolysate_lipid_yield / hydrolysate_lipid_content # 0.343 g microbe / g glucose
hydrolysate_CO2_yield = 1 - hydrolysate_biomass_yield - hydrolysate_lipid_yield # 0.525 g CO2 / g glucose
hydrolysate_biomass_growth_coefficient = hydrolysate_biomass_yield / (hydrolysate_biomass_yield + hydrolysate_CO2_yield) # 0.395 g Biomass / g CO2-Biomass

### Potential range https://onlinelibrary.wiley.com/doi/full/10.1002/ejlt.201100014
max_lipid_yield_glucose = 0.32 # g / g
max_lipid_yield_xylose = 0.34 # g / g
max_titer = 100 # Arbitrary but comparible to ethanol production  [g Lipid / L]
max_productivity = 1.0 # Arbitrary but comparible to ethanol production [g Lipid / L / hr]
min_lipid_yield_glucose = 0.08 # g / g
min_lipid_yield_xylose = 0.08 # g / g
min_titer = 10 # Arbitrary but comparible to baseline  [g Lipid / L]
min_productivity = 0.1 # Arbitrary but comparible to baseline [g Lipid / L / hr]



# -*- coding: utf-8 -*-
"""
"""
import thermosteam as tmo
import biosteam as bst
from biosteam import SystemFactory
from biorefineries.cellulosic import create_hot_water_pretreatment_system
from .bagasse import create_bagasse_drying_system
from .juicing import (
    create_feedstock_handling_system,
    create_juicing_system,
)
from .. import streams as s


@SystemFactory(
    ID='cane_pretreatment_sys',
    ins=[s.cane],
    outs=[s.juice, s.hydrolysate, s.pretreatment_wastewater,
          s.fiber_fines, s.bagasse_to_boiler],
)
def create_cane_combined_1_and_2g_pretreatment(ins, outs):
    """
    Create a system that produces juice and hydrolysate from cane.
    
    """
    oilcane, = ins
    juice, hydrolysate, pretreatment_wastewater, fiber_fines, bagasse_to_boiler = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        ins=oilcane,
        outs='',
        mockup=True,
        area=100
    )
    juicing_sys, udct = create_juicing_system(
        ins=feedstock_handling_sys-0,
        outs=[juice, 'bagasse', fiber_fines],
        mockup=True,
        udct=True,
        area=200,
        dry_bagasse=False,
        pellet_bagasse=False,
    )
    screened_juice, bagasse, fiber_fines = juicing_sys.outs
    
    S1 = bst.Splitter(200, bagasse, split=0.85) # 15% is sent to cogeneration
    S1.isbagasse_splitter = True
    
    create_bagasse_drying_system(ins=S1-1, outs=bagasse_to_boiler, area=200)
    
    udct['S201'].isplit['Lipid'] = 1. # Vibrating screen
    crushing_mill = udct['U201']
    crushing_mill.tag = "oil extraction"
    crushing_mill.isplit['Lipid'] = 0.90
    conveying_belt = S1.outs[0].source
    conveying_belt.cellulose_rxn = tmo.Reaction('Cellulose -> Glucan', 'Cellulose', 1.0, basis='wt')
    conveying_belt.cellulose_rxn.basis = 'mol'
    # Bagasse composition https://www.sciencedirect.com/science/article/pii/S0144861710005072
    # South american; by HPLC
    # Glucan: 41.3%
    # Xylan: 24.9%
    # Galactan: 0.6%
    # Arabinan: 1.7%
    # Lignin: 23.2%
    # Acetyl: 3.0%
    conveying_belt.hemicellulose_rxn = tmo.Reaction('30.2 Hemicellulose -> 24.9 Xylan + 1.7 Arabinan + 0.6 Galactan + 3 Acetate', 'Hemicellulose', 1.0, basis='wt')
    conveying_belt.hemicellulose_rxn.basis = 'mol'
    @conveying_belt.add_specification
    def convert_hemicellulose():
        conveying_belt.run()
        bagasse = conveying_belt.outs[0]
        conveying_belt.cellulose_rxn(bagasse)
        conveying_belt.hemicellulose_rxn(bagasse)
        
    hot_water_pretreatment_sys, hw_dct = create_hot_water_pretreatment_system(
        outs=(hydrolysate, pretreatment_wastewater),
        ins=S1-0,
        mockup=True,
        area=300,
        udct=True,
        solids_loading=0.50, # 50 wt/wt % solids content
    )


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 21:22:25 2022

@author: yrc2
"""
from biorefineries.cane import process_settings

__all__ = process_settings.__all__

from biorefineries.cane.process_settings import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.ethanol import (
    create_beer_distillation_system,
    create_ethanol_purification_system_after_beer_column,
)
from .ethanol import create_sugarcane_to_ethanol_system
from ..fermentation import create_cane_to_combined_1_and_2g_fermentation

__all__ = (
    'create_sugarcane_to_ethanol_combined_1_and_2g',
)

@SystemFactory(
    ID='sugarcane_sys',
    ins=create_sugarcane_to_ethanol_system.ins[:1],
    outs=create_sugarcane_to_ethanol_system.outs[:1],
)
def create_sugarcane_to_ethanol_combined_1_and_2g(ins, outs, WWT_kwargs=None):
    sugarcane, = ins
    ethanol, = outs
    cane_to_fermentation_sys = create_cane_to_combined_1_and_2g_fermentation('cane_to_fermentation_sys', ins=sugarcane)
    beer, lignin, condensate, pretreatment_wastewater, fiber_fines, bagasse_to_boiler = cane_to_fermentation_sys.outs
    cellulosic_beer_distillation_sys = create_beer_distillation_system(
        ins=beer,
        outs=[''],
        mockup=True,
        area=400,
    )
    stripper_process_water = bst.Stream('')
    distilled_beer, stillage = cellulosic_beer_distillation_sys.outs
    ethanol_purification_sys, ep_dct = create_ethanol_purification_system_after_beer_column(
        ins=distilled_beer,
        outs=[ethanol, stripper_process_water],
        mockup=True,
        udct=True,
        area=400,
    )
    C603_3 = bst.SolidsCentrifuge(400, stillage, 
                                  split={'Cellmass': 0.99}, solids=('Cellmass',))
    s = f.stream
    u = f.unit
    bst.Mixer(700, (lignin, C603_3-0, s.filter_cake, bagasse_to_boiler))
    MX = bst.Mixer(400, [condensate, stripper_process_water])
    if WWT_kwargs is None:
        WWT_kwargs = dict(area=500)
    else:
        WWT_kwargs['area'] = 500
    bst.create_all_facilities(
        feedstock=s.bagasse,
        recycle_process_water_streams=[MX-0],
        WWT_kwargs=WWT_kwargs,
        CHP_kwargs=dict(
            ID=700
        ),
        HXN_kwargs=dict(
            ID=1000, 
            ignored=lambda: [u.H401, u.H402],
            Qmin=1e3,
            acceptable_energy_balance_error=0.01
        ),
        area=900,
    )



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 21:33:49 2022

@author: yrc2
"""
from biorefineries.tea import cellulosic_ethanol_tea

__all__ = [*cellulosic_ethanol_tea.__all__, 'create_tea']

from biorefineries.tea.cellulosic_ethanol_tea import *

create_tea = create_cellulosic_ethanol_tea



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>,
#                      Sarang Bhagwat <sarangb2@illinois.edu>,
#                      Yoel Cortes-Pena <yoelcortes@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

import biosteam as bst
from biosteam.process_tools import UnitGroup
from ..cellulosic.systems.pretreatment import create_dilute_acid_pretreatment_system
from . import (
    create_chemicals,
    create_preprocessing_process,
    create_conversion_process,
    create_separation_process,
    create_wastewater_process,
    create_facilities,
    get_splits,
    price,
    set_GWPCF,
    set_FECCF,
    )

__all__ = ('create_system',)


# %%

def create_system(ID='lactic_sys', kind='SSCF', feedstock='feedstock',
                  if_HXN=True, if_BDM=False,
                  flowsheet=None, return_groups=False,
                  cell_mass_split=None, gypsum_split=None,
                  AD_split=None, MB_split=None):
    try: bst.settings.get_chemicals()
    except: create_chemicals()
    kind = kind.upper()
    flowsheet = flowsheet or bst.main_flowsheet
    s = flowsheet.stream
    u = flowsheet.unit

    # Default splits
    _cell_mass_split, _gypsum_split, _AD_split, _MB_split = get_splits()
    cell_mass_split = cell_mass_split if cell_mass_split is not None else _cell_mass_split
    gypsum_split = gypsum_split if gypsum_split is not None else _gypsum_split
    AD_split = AD_split if AD_split is not None else _AD_split
    MB_split = MB_split if MB_split is not None else _MB_split

    create_preprocessing_process(flowsheet=flowsheet, feedstock=feedstock)
    
    # For pretreatment, 93% purity
    sulfuric_acid_T201 = bst.Stream('sulfuric_acid_T201', units='kg/hr',
                                    H2SO4=0.93, H2O=0.07, price=price['H2SO4'])
    set_GWPCF(sulfuric_acid_T201, 'H2SO4')
    set_FECCF(sulfuric_acid_T201, 'H2SO4')    
    
    # For neutralization of pretreatment hydrolysate
    ammonia_M205 = bst.Stream('ammonia_M205', phase='l', units='kg/hr', NH4OH=1, price=price['NH4OH'])
    set_GWPCF(ammonia_M205, 'NH4OH')
    set_FECCF(ammonia_M205, 'NH4OH')
    
    pretreatment_sys = create_dilute_acid_pretreatment_system(
        ins=[u.U101-0, sulfuric_acid_T201, ammonia_M205],
    )
    s.warm_process_water_1.register_alias('water_M201')
    s.warm_process_water_2.register_alias('water_M202')
    s.pretreatment_steam.register_alias('steam_M203')
    s.ammonia_process_water.register_alias('water_M205')
    pretreatment_sys.register_alias('pretreatment_sys')
    
    create_conversion_process(kind=kind, feed=u.P202-0, flowsheet=flowsheet,
                              cell_mass_split=cell_mass_split)
    create_separation_process(feed=u.R301-0, kind=kind,
                              flowsheet=flowsheet,
                              cell_mass_split=cell_mass_split,
                              gypsum_split=gypsum_split)

    if kind == 'SSCF':
        # The last one is reserved for blowdown
        ww_streams = (u.H201-0, u.M401_P-0, u.R402-1, u.R403-1, '')
        solids_to_boiler = (u.U101-1, u.S401-0,)
    elif kind == 'SHF':
        ww_streams = (u.H201-0, u.E301-1, u.M401_P-0, u.R402-1, u.R403-1, '')
        solids_to_boiler = (u.U101-1, u.S301-0, u.S401-0,)
    else: raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')

    create_wastewater_process(ww_streams=ww_streams, flowsheet=flowsheet,
                              AD_split=AD_split, MB_split=MB_split)
    solids_to_boiler = (*solids_to_boiler, u.S504-1)

    process_water_streams = {
        'pretreatment': (s.water_M201, s.water_M202, s.steam_M203, s.water_M205),
        'conversion': (s.water_M301, s.water_R301),
        'separation': (s.water_R403,)
        }
    create_facilities(solids_to_boiler, gas_to_boiler=u.R501-0,
                      treated_water=u.S505-0,
                      process_water_streams=process_water_streams,
                      if_HXN=if_HXN, if_BDM=if_BDM, flowsheet=flowsheet)

    no_hu_facilities = (u.FT, u.PWC, u.ADP, u.CIP)
    if if_HXN: facilities = (u.HXN, u.BT, u.CT, *no_hu_facilities)
    if if_BDM: facilities = (*facilities, u.BDM)

    sys = flowsheet.system
    lactic_sys = bst.System(ID,
                            path=(u.U101, sys.pretreatment_sys, sys.conversion_sys,
                                  sys.separation_sys, sys.wastewater_sys,
                                  u.T601, u.T601_P, u.T602, u.T603, u.T604, u.T605,
                                  u.T606, u.T606_P, u.M601),
                            facilities=facilities,
                            facility_recycle=u.BDM-0 if hasattr(u, 'BDM') else None
                            )
    # lactic_sys.set_tolerance(mol=1e-3, rmol=1e-3, subsystems=True)

    if not return_groups: return lactic_sys

    groups = [
        UnitGroup('preprocessing_group', units=sys.preprocessing_sys.units),
        UnitGroup('pretreatment_group', units=sys.pretreatment_sys.units),
        UnitGroup('conversion_group', units=sys.conversion_sys.units),
        UnitGroup('separation_group', units=sys.separation_sys.units),
        UnitGroup('wastewater_group', units=sys.wastewater_sys.units),
        ]

    if if_HXN: groups.append(UnitGroup('HXN_group', units=(u.HXN,)))
    facility_units = [u.T601, u.T601_P, u.T602, u.T603,
                      u.T604, u.T605, u.T606, u.T606_P, u.M601,
                      *no_hu_facilities]
    groups.extend([
        UnitGroup('BT_group', units=(u.BT,)),
        UnitGroup('CT_group', units=(u.CT,)),
        UnitGroup('facilities_no_hu_group', units=facility_units)
        ])

    return lactic_sys, groups


# -*- coding: utf-8 -*-
"""
"""
from . import chemicals
from . import composition
from . import oil_extraction
from . import process_settings
from . import units
from . import systems
from . import biorefinery
from . import evaluation
from . import results
from . import contour_plots
from . import feature_mockups
from . import uncertainty_plots
from . import parse_configuration
from . import tables

__all__ = (
    *chemicals.__all__,
    *composition.__all__,
    *oil_extraction.__all__,
    *process_settings.__all__,
    *units.__all__,
    *systems.__all__,
    *biorefinery.__all__,
    *evaluation.__all__,
    *results.__all__,
    *contour_plots.__all__,
    *uncertainty_plots.__all__,
    *parse_configuration.__all__,
    *tables.__all__,
)

from .chemicals import *
from .composition import *
from .oil_extraction import *
from .process_settings import *
from .units import *
from .systems import *
from .biorefinery import *
from .evaluation import *
from .results import *
from .contour_plots import *
from .feature_mockups import *
from .uncertainty_plots import *
from .parse_configuration import *
from .tables import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import (units,
               _process_settings,
               _chemicals,
               systems,
               _tea,
)

__all__ = [*_process_settings.__all__,
           *_chemicals.__all__,
           *systems.__all__,
           *_tea.__all__,
           'cornstover_sys',
           'cornstover_tea', 
           'flowsheet',
           'Area100',
           'Area200',
           'Area300',
           'Area400',
           'Area500',
           'Area600',
           'Area700',
           'Area800',
           'AllAreas',
           'areas',
]

from .units import *
from ._process_settings import *
from ._chemicals import *
from .systems import *
from ._tea import *
from biorefineries.cellulosic import Biorefinery

_biorefinery_loaded = False

def load():
    global _biorefinery_loaded
    br = Biorefinery('corn stover ethanol')
    dct = globals()
    dct.update(br.__dict__)
    _biorefinery_loaded = True

def __getattr__(name):
    if not _biorefinery_loaded:
        load()
        dct = globals()
        if name in dct: return dct[name]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 21:27:15 2022

@author: yrc2
"""
from biorefineries.cane import chemicals 

__all__ = [*chemicals.__all__, 'create_chemicals']

from biorefineries.cane.chemicals import *

create_chemicals = create_sugarcane_chemicals


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 13 21:44:23 2019

@author: yoelr
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
from biosteam.plots import (plot_scatter_points, plot_horizontal_line,
                           plot_montecarlo, plot_vertical_line)
import os
from biorefineries.cornstover.model import cornstover_model
import thermosteam as tmo
from thermosteam.utils import colors
from colorpalette import Color
                    
red = Color('red', '#ed5a6a')
red_shade = Color('red_shade', '#d45063')
orange = Color('orange', '#f98f60')
orange_shade = Color('orange', '#de7e55')
green = Color('green', '#7ac083')
green_shade = Color('green_shade', '#6ba873')
blue = Color('blue', '#60c1cf')
blue_shade = Color('blue_shade', '#55a8b5')
purple = Color('purple', '#a381ba')
purple_shade = Color('purple_shade', '#8b6f9d')
black = Color('black', '#3a4551')

# %% Setup of subplots

# light_color = colors.blue_tint.RGBn
# dark_color = colors.blue_shade.RGBn
# dot_color = colors.purple_shade.RGBn
nums = tuple(range(1, 9))
tmo.utils.plots.set_font(size=8)
tmo.utils.plots.set_figure_size(width=6.6142, aspect_ratio=0.45)
fig, axes = plt.subplots(ncols=1, nrows=2, constrained_layout=True, gridspec_kw=dict(height_ratios=[1, 4]))
magnitude_ax, metric_over_benchmark_ax = axes
plt.sca(metric_over_benchmark_ax)

# %% Plot MESP and ethanol sales

positions_other = (16, 17, 18)
other_index = [('Biorefinery', 'Steam demand [kg/hr]'),
               ('Biorefinery', 'Ethanol production [kg/hr]'),
               ('Biorefinery', 'Minimum ethanol selling price [USD/gal]')]
data = cornstover_model.metrics_at_baseline()
other_data = np.array(data[other_index])
other_data[0] *= 100./234784.
other_data[1] *= 100./22273.
other_data[2] *= 100./2.15
bx_other = plt.bar(positions_other, other_data, 0.5,
        align='center',
        color=blue.RGBn,
        edgecolor=blue_shade.RGBn)

# %% Plot electricity

plot_vertical_line(15.5, color=colors.grey_tint.RGBn)
positions_electricity = tuple(range(9))
areas = [f'Area {i}00' for i in nums]
units = 'MW'
positions = np.arange(0, 9)
electricity_cols = [(i, 'Electricity [MW]') for i in areas]
humbird_electricity = 41 * np.array((0.02, 0.14, 0.06, 0.05,
                                     0.18, 0.003, 0.03, 0.08,
                                     0.44))
electricity_data = data[electricity_cols]  #/ humbird_electricity
electricity_data[('Biorefinery', 'Excess electricity [MW]')] = data[('Biorefinery', 'Excess electricity [MW]')]
electricity_data_humbird_normalized = electricity_data * (100/humbird_electricity)
bx_electricity = plt.bar(positions_electricity, electricity_data_humbird_normalized, 0.5,
                         align='center',
                         color=orange.RGBn,
                         edgecolor=orange_shade.RGBn)

# plot_vertical_line(7.5, color=colors.orange_tint.shade(15).RGBn, ls='-.')
# plot_vertical_line(9.5, color=colors.orange_tint.shade(15).RGBn, ls='-.')

# %% Plot installed cost

units = '10^6 USD'
plot_vertical_line(8.5, color=colors.grey_tint.RGBn)
positions_installed = tuple(range(9, 16))
installed_cols = [(i, 'Installed equipment cost [10^6 USD]') for i in areas[1:]]
humbird_installed = np.array([24.2, 32.9, 31.2, 22.3, 49.4, 5, 66, 6.9])
installed_data = data[installed_cols]
# installed_data[('Biorefinery', 'Installed cost')] = installed_data.sum(1)
installed_data_humbird_normalized = installed_data * (100/humbird_installed[1:])
bx_installed = plt.bar(
    positions_installed, installed_data_humbird_normalized, 0.5,
    align='center',
    color=red.RGBn,
    edgecolor=red_shade.RGBn,
)
plot_horizontal_line(100, ls='--')
u_lb = 0; y_ub = 200
plt.ylim(0, 200)
yticks = np.arange(0, 201, 50)
plt.yticks(yticks)
y_text = 0.885*y_ub
y_letter = 0.875 * y_ub
plt.text(4, y_text, "Electricity demand", color=orange_shade.RGBn,
          horizontalalignment='center', fontsize=12, fontweight='bold')
plt.text(-0.25, y_letter, "C", color=black.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.text(12, y_text, "Installed cost", color=red_shade.RGBn,
          horizontalalignment='center', fontsize=12, fontweight='bold')
plt.text(8.75, y_letter, "D", color=black.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.text(15.75, y_letter, "E", color=black.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
# plt.text(9.25, y_letter, "D", color=black.RGBn,
#          horizontalalignment='center', fontsize=14, fontweight='bold')

# plt.text(16, y_text, "Ethanol\nsales", color=colors.red_shade.RGBn,
#           horizontalalignment='center', fontsize=12, fontweight='bold')
# plt.text(16.0, y_text, "MESP", color=colors.blue_shade.RGBn,
#           horizontalalignment='center', fontsize=12, fontweight='bold')

plt.xlim(-0.5, 18.5)
plt.ylabel("Metric over benchmark [%]")
area_marks = [i.replace(' ', '\n') for i in areas]
xmarks = area_marks + ['Excess'] + area_marks[1:] + ['Steam\ndemand', '   EtOH\n    prod.', '    MESP']
xticks = positions_electricity + positions_installed + positions_other
plt.xticks(xticks, xmarks)
metric_over_benchmark_ax.set_zorder(1e6)

plt.sca(magnitude_ax)
plt.fill_between([-0.5, 15.5], 0, 1, color=colors.neutral_tint.tint(85).RGBn)

electricity_areas = humbird_electricity # electricity_data.median()
electricity_areas /= max(electricity_areas)
plt.bar(positions_electricity, electricity_areas, 0.5,
        align='center', label="Electricity demand",
        color=orange_shade.RGBn,
        edgecolor=orange_shade.tint(10).shade(15).RGBn)
plot_vertical_line(8.5, color=colors.grey_tint.RGBn)

installed_areas = humbird_installed[1:]# installed_data.median()
installed_areas /= max(installed_areas)
plt.bar(positions_installed, installed_areas, 0.5,
        align='center', label="Installed equipment cost",
        color=red_shade.RGBn,
        edgecolor=red_shade.tint(10).shade(15).RGBn)

plot_vertical_line(15.5, color=colors.grey_tint.RGBn)
# plot_vertical_line(15.5, color='k', lw=0.8)
# plt.bar(positions_MESP, [1], 0.5,
#         align='center', label="MESP",
#         color=colors.blue.tint(30).shade(15).RGBn,
#         edgecolor=colors.blue_shade.RGBn)

plot_vertical_line(-0.5, color='k')
plt.hlines([1], [-0.5], [15.5], color='k')
magnitude_ax.spines['top'].set_visible(False)
magnitude_ax.spines['right'].set_visible(False)
magnitude_ax.tick_params(axis="x", direction="in", length=0)
magnitude_ax.set_zorder(2)
metric_over_benchmark_ax.set_zorder(1)
plt.yticks([], [])
plt.xticks(xticks[:-3], [])
plt.ylim(0, 1)
plt.xlim(-0.5, 18.5)
plt.text(4, 0.65, "Benchmark magnitude", color=black.RGBn,
         horizontalalignment='center', fontsize=12, fontweight='bold')
plt.text(-0.25, 0.60, "A", color=black.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.text(8.75, 0.60, "B", color=black.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.subplots_adjust(hspace=.0)

metric_over_benchmark_ax.tick_params(axis='x', direction="inout", length=4)
for ax in axes:
    ax.tick_params(axis='y', right=False, direction="inout", length=4)
ax2 = metric_over_benchmark_ax.twinx()
plt.sca(ax2)
plt.ylim(0, 200)
plt.yticks(yticks, [])
ax2.zorder = 1000
ax2.tick_params(direction="in")
    
xlabels = metric_over_benchmark_ax.get_xticklabels()    

plt.subplots_adjust(
    hspace=0, wspace=0.05,
    top=0.95, bottom=0.15,
    left=0.08, right=0.98,
)

images_folder = os.path.join(os.path.dirname(__file__), 'images')
for i in ('svg', 'png'):
    file = os.path.join(images_folder, f'single_point_evaluation.{i}')
    plt.savefig(file, transparent=True)


# -*- coding: utf-8 -*-
"""
Created on Fri May 14 22:48:34 2021

@author: yrc2
"""
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import os
from chaospy import distributions as shape
from math import sqrt

__all__ = (
    'ethanol_no_RIN_price_distribution', 
    'advanced_ethanol_price_distribution',
    'cellulosic_ethanol_price_distribution',
    'biomass_based_diesel_price_distribution',
    'cellulosic_based_diesel_price_distribution',
    'natural_gas_price_distribution',
    'mean_glycerol_price',
    'mean_ethanol_no_RIN_price',
    'mean_advanced_ethanol_price',
    'mean_cellulosic_ethanol_price',
    'mean_biomass_based_diesel_price',
    'mean_cellulosic_based_diesel_price',
    'mean_natural_gas_price',
    'mean_electricity_price',
    'mean_soymeal_price',
)

liter_per_gal = 3.7854
RINfile = os.path.join(os.path.dirname(__file__), 'RIN_prices_2022.csv')
df_RIN = pd.read_csv(RINfile)
df_D3 = df_RIN[df_RIN['Fuel (D Code)'] == 'D3'] # Cellulosic ethanol
df_D4 = df_RIN[df_RIN['Fuel (D Code)'] == 'D4'] # Biodiesel
df_D5 = df_RIN[df_RIN['Fuel (D Code)'] == 'D5'] # Advanced ethanol
df_D6 = df_RIN[df_RIN['Fuel (D Code)'] == 'D6'] # Conventional ethanol

def summarize_RIN_credit_by_year(df, month_offset):
    data = {}
    for date, price in zip(df['Transfer Date by Week'], df['RIN Price']):
        month, day, year = date.split('/')
        year = int(year)
        if year in data:
            data_by_year = data[year]
        else:
            data_by_year = data[year] = []
        if int(month) - month_offset < 0: year -= 1
        data_by_year.append(float(price.strip('$')))
    
    for year, data_by_year in data.items():
        data[year] = np.mean(data_by_year)
    return data

def get_RIN_credit_by_year_June2017_to_May2022(df):
    offset = 6
    dct = summarize_RIN_credit_by_year(df, offset)
    years = (2017, 2018, 2019, 2020, 2021)
    return np.array([dct[y] for y in years])

# June 2017 to May 2022 (start of year is offset by 6 months)
RIN_D3_prices = get_RIN_credit_by_year_June2017_to_May2022(df_D3) / liter_per_gal
RIN_D4_prices = get_RIN_credit_by_year_June2017_to_May2022(df_D4) / liter_per_gal
RIN_D5_prices = get_RIN_credit_by_year_June2017_to_May2022(df_D5) / liter_per_gal
RIN_D6_prices = get_RIN_credit_by_year_June2017_to_May2022(df_D6) / liter_per_gal


# %% Raw data

ethanol_no_RIN_prices = np.array([ # USD / gal May 2017 - Dec 2020, by month
    1.6, 1.62, 1.64, 1.55, 1.19, 0.98, 0.95, 1.0, 1.11, 1.25, 1.31,
    1.34, 1.4, 1.41, 1.31, 1.2, 1.21, 1.19, 1.08, 1.09, 1.12, 1.18, 1.17,
    1.21, 1.36, 1.31, 1.3, 1.22, 1.39, 1.44, 1.32, 1.15, 1.03, 0.84, 0.59,
    0.75, 0.85, 0.91, 0.87, 0.95, 0.94, 0.78, 0.6, 
]) / liter_per_gal

ethanol_with_D6RIN_prices = np.array([ # USD / gal Jan 2021  - May 2022, by month
    1.55, 1.73, 1.87, 2.15, 2.62, 2.53, 2.37, 2.35, 2.52, 2.62, 3.42, 3.11, 
    2.22, 2.17, 2.49, 2.64, 2.82
]) / liter_per_gal

assert len(ethanol_no_RIN_prices) + len(ethanol_with_D6RIN_prices) == 5 * 12

# Includes RIN D4
biomass_based_diesel_prices = np.array([ # USD / gal June 2017 - June 2022, by month
    3.13, 3.23, 3.31, 3.33, 3.3, 3.34, 3.23, 3.45, 3.16, 3.04, 3.02, 3.13,
    3.11, 3.02, 3.01, 3.04, 3.09, 3.09, 3.06, 3.12, 3.24, 3.08, 2.98, 2.86,
    2.87, 2.91, 2.99, 3.06, 3.13, 3.21, 3.26, 3.35, 3.11, 2.97, 2.77, 2.74, 
    2.87, 2.92, 3.12, 3.3, 2.88, 3.51, 3.61, 4.08, 4.5, 4.91, 5.0, 5.5, 5.99, 
    5.95, 5.91, 5.71, 5.6, 5.49, 4.88, 4.83, 6.03, 6.5, 6.99, 7.53, 7.35
]) / liter_per_gal

def by_month_to_year(arr):
    N_years = int(len(arr)/12)
    return np.array([ 
        np.mean(arr[12*i: 12*i+12]) for i in range(N_years)
    ])

biomass_based_diesel_prices = by_month_to_year(biomass_based_diesel_prices)
biodiesel_no_RIN_prices = biomass_based_diesel_prices - RIN_D4_prices * 1.5
cellulosic_based_diesel_prices = biodiesel_no_RIN_prices + RIN_D3_prices * 1.5
ethanol_no_RIN_prices = by_month_to_year(ethanol_no_RIN_prices)
ethanol_with_D6RIN_prices = by_month_to_year(ethanol_with_D6RIN_prices)
ethanol_no_RIN_prices = np.array([
    *ethanol_no_RIN_prices, 
    *(ethanol_with_D6RIN_prices - RIN_D6_prices[-2:]),
])
assert len(ethanol_no_RIN_prices) == 5

natural_gas_prices = np.array([ # City gate [USD / mcf] 2017 to 2021, mcf = 1,000 cf
    4.16, 4.23, 3.81, 3.43, 6.11
]) * 35.3146667/1e3 # To USD / m3

crude_oil_prices = np.array([ # USD / barrel West Texas Intermediate 2018 - 2022; https://www.macrotrends.net/1369/crude-oil-price-history-chart
    65.23, 56.99, 39.68, 68.17, 94.53
])

# %% Shorten data to 2017 - 2022 because of data limitations for cellulosic ethanol RINs 

# Account for RIN credits
cellulosic_ethanol_prices = ethanol_no_RIN_prices + RIN_D3_prices
advanced_ethanol_prices = ethanol_no_RIN_prices + RIN_D5_prices

# plt.plot(cellulosic_ethanol_prices)
# plt.plot(advanced_ethanol_prices)
# plt.plot(biomass_based_diesel_prices)

def triangular_distribution(x, median=False):
    a, b, c = fit_triangular_distribution(x, median)
    return shape.Triangle(a, c, b)

def triangular_distribution_median(a, b, c):
    if c >= (a + b) / 2: 
        return a + sqrt((b - a) * (c - a) * 0.5)
    else:
        return b - sqrt((b - a) * (b - c) * 0.5)

def fit_triangular_distribution(x, median=False):
    a = min(x)
    b = max(x)
    if median:
        median = np.median(x)
        if median >= (a + b) / 2: 
            dummy = a - median
            c = a + 2 * dummy * dummy / (b - a)
        else:
            dummy = b - median
            c = b - 2 * dummy * dummy / (b - a)
        assert triangular_distribution_median(a, b, c) == median 
    else:
        c = 3 * np.mean(x) - (a + b)
        if c < a: c = a + 1e-6
    return a, b, c

def plot_triangular_distribution(a, b, c):
    return plt.plot([a, b, c], [0, 2 / (b - a), 1])

def plot_histogram(x, *args, bins=10, density=True, **kwargs):
    return plt.hist(x, *args, **kwargs)

# Price distributions
electricity_price_distribution = shape.Triangle(0.0583, 0.065, 0.069) # Data from historical prices, 2010-2020; https://www.eia.gov/outlooks/aeo/pdf/00%20AEO2021%20Chart%20Library.pdf
ethanol_no_RIN_price_distribution = triangular_distribution(ethanol_no_RIN_prices)
advanced_ethanol_price_distribution = triangular_distribution(advanced_ethanol_prices)
cellulosic_ethanol_price_distribution = triangular_distribution(cellulosic_ethanol_prices)
biomass_based_diesel_price_distribution = triangular_distribution(biomass_based_diesel_prices)
cellulosic_based_diesel_price_distribution = triangular_distribution(cellulosic_based_diesel_prices)
natural_gas_price_distribution = triangular_distribution(natural_gas_prices)
crude_oil_price_distribution = triangular_distribution(crude_oil_prices)

# https://www.eia.gov/outlooks/aeo/pdf/00%20AEO2021%20Chart%20Library.pdf
# Data from historical prices, 2010-2020
electricity_price_distribution = shape.Triangle(0.0583, 0.065, 0.069)

# Mean values
mean_glycerol_price = (0.10 + 0.22) * 0.5 
mean_RIN_D3_price = np.mean(RIN_D3_prices)
mean_RIN_D5_price = np.mean(RIN_D5_prices)
mean_ethanol_no_RIN_price = np.mean(ethanol_no_RIN_prices)
mean_cellulosic_ethanol_price = np.mean(cellulosic_ethanol_prices)
mean_advanced_ethanol_price = np.mean(advanced_ethanol_prices)
mean_biomass_based_diesel_price = np.mean(biomass_based_diesel_prices)
mean_cellulosic_based_diesel_price = np.mean(cellulosic_based_diesel_prices)
mean_natural_gas_price = np.mean(natural_gas_prices)
mean_electricity_price = sum([0.0583, 0.065, 0.069]) / 3.
mean_soymeal_price = 0.33 # 10 yr average; https://markets.businessinsider.com/commodities/soybean-meal-price?op=1
mean_crude_oil_price = np.mean(crude_oil_prices)
mcop = mean_crude_oil_price

# Short hand distributions names
cbpd = cellulosic_based_diesel_price_distribution 
mcbp = mean_cellulosic_based_diesel_price
bpd = biomass_based_diesel_price_distribution 
mbp = mean_biomass_based_diesel_price
cepd = cellulosic_ethanol_price_distribution 
mcep = mean_cellulosic_ethanol_price
aepd = advanced_ethanol_price_distribution 
maep = mean_advanced_ethanol_price
copd = crude_oil_price_distribution


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""
from biosteam import stream_kwargs

vegetable_oil = stream_kwargs(
    'vegetable_oil',
    Water=0.0184,
    TAG=11.1
)
biodiesel = stream_kwargs('biodiesel', price=1.38) # assuming density = 870 kg/m3
crude_glycerol = stream_kwargs('crude_glycerol')
wastewater = stream_kwargs('wastewater', price=-0.33)

crude_vegetable_oil = stream_kwargs(
    'crude_vegetable_oil',
    Water=0.0184,
    TAG=11.1,
    PL=0.1
)
acetone = stream_kwargs('acetone', price=0.80)
pure_glycerine = stream_kwargs('pure_glycerine', price=0.65)
degummed_oil = stream_kwargs('degummed_oil')
polar_lipids = stream_kwargs('polar_lipids')
wastewater = stream_kwargs('wastewater')
lipid = stream_kwargs('lipid')
lipid_wash_water = stream_kwargs('lipid_wash_water')
washed_lipid = stream_kwargs('washed_lipid')
spent_wash_water = stream_kwargs('spent_wash_water')


# -*- coding: utf-8 -*-
"""
"""
from biosteam import stream_kwargs
from biorefineries.cellulosic.streams import *
from biorefineries.ethanol.streams import *
from biorefineries.biodiesel.streams import *

bagasse = stream_kwargs('bagasse')
bagasse_pellets = stream_kwargs('bagasse_pellets')
sugarcane = stream_kwargs('sugarcane',
    Water=0.7,
    Glucose=0.01208,
    Sucrose=0.1369,
    Ash=0.006,
    Cellulose=0.06115,
    Hemicellulose=0.03608,
    Lignin=0.03276,
    Solids=0.015,
    total_flow=333334.2,
    units='kg/hr',
    price=0.03455
)
lipidcane = stream_kwargs('lipidcane',
    Ash=2000.042,
    Cellulose=26986.69,
    Glucose=2007.067,
    Hemicellulose=15922.734,
    Lignin=14459.241,
    TAG=10035.334,
    Solids=5017.667,
    Sucrose=22746.761,
    Water=234157.798,
    units='kg/hr',
    price=0.03455,
)
oilcane = lipidcane.copy()
oilcane['ID'] = 'oilcane'
TAG = oilcane['TAG']
oilcane['TAG'] = 0.80 * TAG
oilcane['PL'] = 0.10 * TAG
oilcane['FFA'] = 0.10 * TAG
cane = stream_kwargs('cane')
condensate = stream_kwargs('condensate')
bagasse_to_boiler = stream_kwargs('bagasse_to_boiler')
shredded_cane = stream_kwargs('shredded_cane')
untreated_juice = stream_kwargs('untreated_juice')
H3PO4 = stream_kwargs('H3PO4',
    H3PO4=74.23,
    Water=13.1,
    units='kg/hr',
    price=0
)
lime = stream_kwargs('lime',
    CaO=333.0,
    Water=2200.0,
    units='kg/hr',
    price=0.077
)
polymer = stream_kwargs('polymer',
    Flocculant=0.83,
    units='kg/hr',
    price=0
)
clarified_juice = stream_kwargs('clarified_juice')
screened_juice = stream_kwargs('screened_juice', 
    Glucose=3802,
    Sucrose=4.309e+04,
    Water=2.59e+05,
    H3PO4=83.33,
    units='kg/hr',
    T=372
)
stillage = stream_kwargs('stillage')
fiber_fines = stream_kwargs('fiber_fines')
evaporator_condensate = stream_kwargs('evaporator_condensate')
vent = stream_kwargs('vent')
vinasse = stream_kwargs('vinasse')
wastewater = stream_kwargs('wastewater')
emissions = stream_kwargs('emissions')
ash_disposal = stream_kwargs('ash_disposal')
molasses = stream_kwargs('molasses')
lipid = stream_kwargs('lipid')
cellmass = stream_kwargs('cellmass')
fermentation_effluent = stream_kwargs('fermentation_effluent')
spent_oil_wash_water = stream_kwargs('spent_oil_wash_water')
sugar = stream_kwargs('sugar', price=0.419) # https://markets.businessinsider.com/commodities/sugar-price?op=1 (3/18/2022)
acTAG = stream_kwargs('acTAG', price=1.633) 
crude_oil = stream_kwargs('crude_oil', price=0.661) # 30 cts / lb (vegetable); 5yr average https://www.ams.usda.gov/mnreports/lswagenergy.pdf


# -*- coding: utf-8 -*-
"""
Created on Thu Nov  4 14:44:17 2021

@author: yrc2
"""
from biosteam import MockFeature

(set_juicing_oil_recovery, set_microbial_oil_recovery, set_bagasse_oil_recovery, 
 set_cane_operating_days, set_sorghum_operating_days, 
 set_available_land, set_crude_oil_price, set_baseline_feedstock_price,
 set_cellulosic_ethanol_price, set_advanced_ethanol_price,
 set_biomass_based_diesel_price, set_cellulosic_based_diesel_price,
 set_natural_gas_price, set_electricity_price, 
 set_IRR, set_crude_glycerol_price, set_pure_glycerol_price,
 set_saccharification_reaction_time, set_cellulase_price, 
 set_cellulase_loading, set_reactor_base_cost,
 set_cane_glucose_yield, set_cane_xylose_yield, 
 set_sorghum_glucose_yield, set_sorghum_xylose_yield, 
 set_glucose_to_ethanol_yield, set_xylose_to_ethanol_yield,
 set_cofermentation_ethanol_titer, set_cofermentation_ethanol_productivity,
 set_glucose_to_microbial_oil_yield, set_xylose_to_microbial_oil_yield,
 set_fermentation_microbial_oil_titer, set_fermentation_microbial_oil_productivity,
 set_cane_PL_content, set_sorghum_PL_content, set_cane_FFA_content,
 set_sorghum_FFA_content,  set_cane_oil_content, set_relative_sorghum_oil_content,
 set_TAG_to_FFA_conversion, set_feedstock_GWP, set_methanol_GWP, 
 set_pure_glycerine_GWP, set_cellulase_GWP, set_natural_gas_GWP,
 set_income_tax,
) = all_parameter_mockups = (
    MockFeature('Juicing oil recovery', '%', ''),
    MockFeature('Microbial oil recovery', '%', ''),
    MockFeature('Bagasse oil recovery', '%', ''),
    MockFeature('Cane operating days', 'day/yr', ''),
    MockFeature('Sorghum operating days', 'day/yr', ''),
    MockFeature('Available land', 'ha / yr', ''),
    MockFeature('Price', 'USD/barrel', 'Crude oil'),
    MockFeature('Price', 'USD/kg', 'Feedstock'),
    MockFeature('Price', 'USD/L', 'Cellulosic ethanol'),
    MockFeature('Price', 'USD/L', 'Advanced ethanol'),
    MockFeature('Price', 'USD/L', 'Biomass based diesel'),
    MockFeature('Price', 'USD/L', 'Cellulosic based diesel'),
    MockFeature('Price', 'USD/m3', 'Natural gas'),
    MockFeature('Price', 'USD/kWhr', 'Electricity'),
    MockFeature('IRR', '%', ''),
    MockFeature('Price', 'USD/kg', 'Crude glycerol'),
    MockFeature('Price', 'USD/kg', 'Pure glycerine'),
    MockFeature('Reaction time', 'hr', 'Saccharification'),
    MockFeature('Price', 'USD/kg', 'Cellulase'),
    MockFeature('Cellulase loading', 'wt. % cellulose', 'Cellulase'),
    MockFeature('Base cost', 'million USD', 'Pretreatment reactor system'),
    MockFeature('Cane glucose yield', '%', 'Pretreatment and saccharification'),
    MockFeature('Sorghum glucose yield', '%', 'Pretreatment and saccharification'),
    MockFeature('Cane xylose yield', '%', 'Pretreatment and saccharification'),
    MockFeature('Sorghum xylose yield', '%', 'Pretreatment and saccharification'),
    MockFeature('Glucose to ethanol yield', '%', 'Cofermenation'),
    MockFeature('Xylose to ethanol yield', '%', 'Cofermenation'),
    MockFeature('Ethanol titer', 'g/L', 'Cofermentation'),
    MockFeature('Ethanol productivity', 'g/L', 'Cofermentation'),
    MockFeature('Glucose to microbial oil yield', '%', 'Cofermenation'),
    MockFeature('Xylose to microbial oil yield', '%', 'Cofermenation'),
    MockFeature('Microbial oil titer', 'g/L', 'Fermentation'),
    MockFeature('Microbial oil productivity', 'g/L', 'Fermentation'),
    MockFeature('Cane PL content', '% oil', 'Oilcane'),
    MockFeature('Sorghum PL content', '% oil', 'Oilsorghum'),
    MockFeature('Cane FFA content', '% oil', 'Oilcane'),
    MockFeature('Sorghum FFA content', '% oil', 'Oilsorghum'),
    MockFeature('Cane oil content', 'dry wt. %', 'Oilcane'),
    MockFeature('Relative sorghum oil content', 'dry wt. %', 'Oilsorghum'),
    MockFeature('TAG to FFA conversion', '% oil', ''),
    MockFeature('GWP', 'kg*CO2-eq/kg', 'Sugarcane'),
    MockFeature('GWP', 'kg*CO2-eq/kg', 'Methanol'),
    MockFeature('GWP', 'kg*CO2-eq/kg', 'Pure glycerine'),
    MockFeature('GWP', 'kg*CO2-eq/kg', 'Cellulase'),
    MockFeature('GWP', 'kg*CO2-eq/kg', 'Natural gas'),
    MockFeature('Income tax', '%', ''),
)
     
(MFPP, MESP, MBSP, feedstock_consumption, biodiesel_production, biodiesel_yield, ethanol_production, 
 electricity_production, net_energy_production, natural_gas_consumption, TCI, 
 heat_exchanger_network_error, GWP_economic, GWP_ethanol, GWP_biodiesel, 
 GWP_crude_glycerol, GWP_electricity, GWP_ethanol_displacement, GWP_biodiesel_displacement,
 GWP_biofuel_allocation, GWP_ethanol_allocation,
 GWP_biodiesel_allocation, GWP_crude_glycerol_allocation,
 IRR, MFPP_derivative, 
 biodiesel_production_derivative, ethanol_production_derivative, 
 electricity_production_derivative, natural_gas_consumption_derivative, 
 TCI_derivative, GWP_economic_derivative, 
 GWP_ethanol_derivative, GWP_biodiesel_derivative,
 GWP_crude_glycerol_derivative, GWP_electricity_derivative,
 ROI, competitive_biomass_yield, energy_competitive_biomass_yield,
 # competitive_microbial_oil_yield,
 # energy_competitive_microbial_oil_yield,
 ) = all_metric_mockups = (
    MockFeature('MFPP', 'USD/MT', ''),
    MockFeature('MESP', 'USD/L', ''),
    MockFeature('MBSP', 'USD/L', ''),
    MockFeature('Feedstock consumption', 'MT/yr', ''),
    MockFeature('Biodiesel production', 'L/MT', ''),
    MockFeature('Biodiesel yield', 'L/ha', ''),
    MockFeature('Ethanol production', 'L/MT', ''),
    MockFeature('Electricity production', 'kWhr/MT', ''),
    MockFeature('Net energy production', 'GGE/MT', ''),
    MockFeature('Natural gas consumption', 'm3/MT', ''),
    MockFeature('TCI', '10^6*USD', ''),
    MockFeature('Heat exchanger network error', '%', ''),
    MockFeature('GWP', 'kg*CO2*eq / USD', 'Economic allocation'),
    MockFeature('Ethanol GWP', 'kg*CO2*eq / L', 'Economic allocation'),
    MockFeature('Biodiesel GWP', 'kg*CO2*eq / L', 'Economic allocation'),
    MockFeature('Crude glycerol GWP', 'kg*CO2*eq / kg', 'Economic allocation'),
    MockFeature('Electricity GWP', 'kg*CO2*eq / MWhr', 'Economic allocation'),
    MockFeature('Ethanol GWP', 'kg*CO2*eq / L', 'Displacement allocation'),
    MockFeature('Biodiesel GWP', 'kg*CO2*eq / L', 'Displacement allocation'),
    MockFeature('Biofuel GWP', 'kg*CO2*eq / GGE', 'Energy allocation'),
    MockFeature('Ethanol GWP', 'kg*CO2*eq / L', 'Energy allocation'),
    MockFeature('Biodiesel GWP', 'kg*CO2*eq / L', 'Energy allocation'),
    MockFeature('Crude-glycerol GWP', 'kg*CO2*eq / kg', 'Energy allocation'),
    MockFeature('IRR', '%', ''),
    MockFeature('MFPP derivative', 'USD/MT', ''),
    MockFeature('Biodiesel production derivative', 'L/MT', ''),
    MockFeature('Ethanol production derivative', 'L/MT', ''),
    MockFeature('Electricity production derivative', 'kWhr/MT', ''),
    MockFeature('Natural gas consumption derivative', 'cf/MT', ''),
    MockFeature('TCI derivative', '10^6*USD', ''),
    MockFeature('GWP derivative', 'kg*CO2*eq / USD', 'Economic allocation'),
    MockFeature('GWP derivative', 'kg*CO2*eq / L', 'Ethanol'),
    MockFeature('GWP derivative', 'kg*CO2*eq / L', 'Biodiesel'),
    MockFeature('GWP derivative', 'kg*CO2*eq / kg', 'Crude glycerol'),
    MockFeature('GWP derivative', 'kg*CO2*eq / MWhr', 'Electricity'),
    MockFeature('ROI', '%', ''),
    MockFeature('Competitive biomass yield', 'dry MT/ha', 'Feedstock'),
    MockFeature('Energy competitive biomass yield', 'dry MT/ha', 'Feedstock'),
)

tea_monte_carlo_metric_mockups = (
    MFPP, 
    TCI,
    ethanol_production,
    biodiesel_production,
    electricity_production,
    natural_gas_consumption
)

tea_monte_carlo_derivative_metric_mockups = (
    MFPP_derivative, 
    TCI_derivative,
    ethanol_production_derivative,
    biodiesel_production_derivative,
    electricity_production_derivative,
    natural_gas_consumption_derivative,
)

lca_monte_carlo_metric_mockups = (
    GWP_economic,
    GWP_ethanol,
    GWP_biodiesel,
    GWP_electricity,
    GWP_crude_glycerol,
)

lca_monte_carlo_derivative_metric_mockups = (
    GWP_economic_derivative,
    GWP_ethanol_derivative,
    GWP_biodiesel_derivative, 
    GWP_electricity_derivative,
    GWP_crude_glycerol_derivative,
)


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""

import biosteam as bst
from biorefineries.cellulosic import units
from biorefineries.cellulosic import streams as s
from .integrated_bioprocess import create_integrated_bioprocess_saccharification_and_cofermentation_system
from .cofermentation import create_cofermentation_system
from .saccharification import create_saccharification_system
from .simultaneous_saccharification_cofermentation import create_simultaneous_saccharification_and_cofermentation_system

__all__ = (
    'create_cellulosic_fermentation_system',
)

@bst.SystemFactory(
    ID='cellulosic_fermentation_sys',
    ins=[s.pretreated_biomass, s.cellulase, s.saccharification_water, s.DAP, s.CSL],
    outs=[s.vent, s.beer, s.lignin],
)
def create_cellulosic_fermentation_system(
        ins, outs,
        include_scrubber=None,
        solids_loading=None,
        insoluble_solids_loading=None,
        nonsolids=None,
        insoluble_solids=None,
        kind=None, 
        # Valid arguments include:
        # Integrated Bioprocess (IB), 
        # Simultaneous Saccharification and Co-Fermentation (SSCF),
        # Saccharification and Co-Fermentation (SCF),
        Saccharification=None,
        ContinuousPresaccharification=None,
        SeedTrain=None,
        CoFermentation=None,
        SaccharificationAndCoFermentation=None,
        saccharification_reactions=None,
        seed_train_reactions=None,
        cofermentation_reactions=None,
        add_nutrients=True,
    ):
    vent, beer, lignin = outs
    pretreated_biomass, cellulase, saccharification_water, DAP, CSL = ins
    if not add_nutrients:
        ins.remove(CSL)
        ins.remove(DAP)
    if kind is None: kind = 'IB'
    SCF_keys = ('SCF', 'Saccharification and Co-Fermentation')
    saccharification_sys = create_saccharification_system(
        ins=[pretreated_biomass, cellulase, saccharification_water],
        mockup=True,
        solids_loading=solids_loading,
        insoluble_solids_loading=insoluble_solids_loading,
        nonsolids=nonsolids,
        insoluble_solids=insoluble_solids,
        Saccharification=(Saccharification or units.Saccharification if kind in SCF_keys else ContinuousPresaccharification or units.ContinuousPresaccharification),
        saccharification_reactions=saccharification_reactions,
    )
    if kind in ('IB', 'Integrated Bioprocess'):
        outs.remove(lignin)
        create_integrated_bioprocess_saccharification_and_cofermentation_system(
            ins=[saccharification_sys-0, DAP, CSL],
            outs=[vent, beer],
            mockup=True,
            SaccharificationAndCoFermentation=SaccharificationAndCoFermentation,
            SeedTrain=SeedTrain,
            include_scrubber=include_scrubber,
            seed_train_reactions=seed_train_reactions,
            cofermentation_reactions=cofermentation_reactions,
            add_nutrients=add_nutrients,
        )
    elif kind in ('SSCF', 'Simultaneous Saccharification and Co-Fermentation'):
        outs.remove(lignin)
        create_simultaneous_saccharification_and_cofermentation_system(
            ins=[saccharification_sys-0, DAP, CSL],
            outs=[vent, beer],
            mockup=True,
            include_scrubber=include_scrubber,
            seed_train_reactions=seed_train_reactions,
            cofermentation_reactions=cofermentation_reactions,
            add_nutrients=add_nutrients,
        )
    elif kind in SCF_keys:
        T303 = bst.StorageTank('T303', saccharification_sys-0, tau=4)
        create_cofermentation_system(
            ins=[T303-0, DAP, CSL],
            outs=[vent, beer, lignin],
            mockup=True,
            SeedTrain=SeedTrain,
            CoFermentation=CoFermentation,
            include_scrubber=include_scrubber,
            seed_train_reactions=seed_train_reactions,
            cofermentation_reactions=cofermentation_reactions,
            add_nutrients=add_nutrients,
        )
    else:
        raise ValueError("invalid 'kind'")




# -*- coding: utf-8 -*-
"""
Created on Mon Dec 12 16:21:58 2022

@author: yrc2
"""
from biosteam import stream_kwargs

__all__ = (
    'beer', 'distilled_beer', 'stillage', 'denaturant', 'ethanol', 
    'recycle_process_water',
)

beer = stream_kwargs('beer',
    T=348.34,
    P=101325,
    Water=96500.0,
    Ethanol=22550.0,
    Glucose=4916,
    H3PO4=83.33,
    Yeast=103,
    units='kg/hr'
)
distilled_beer = stream_kwargs('distilled_beer')
stillage = stream_kwargs('stillage')
denaturant = stream_kwargs('denaturant',
    Octane=230.69,
    units='kg/hr',
    price=0.756
)
ethanol = stream_kwargs('ethanol',
    price=0.789
)
recycle_process_water = stream_kwargs('recycle_process_water')

